"""
Robot dynamics operations.

@author: Peter Corke
@copyright: Peter Corke
"""

from numpy import *
from .utility import *
from .transform import *
from . import jacobian as Jac
from numpy.linalg import inv

def accel(robot, *args):
    """
    Compute manipulator forward dynamics, the joint accelerations that result
    from applying the actuator torque to the manipulator robot in state q and qd.

       - qdd = accel(robot, q, qd, torque)
       - qdd = accel(robot, [q qd torque])

    Uses the method 1 of Walker and Orin to compute the forward dynamics.
    This form is useful for simulation of manipulator dynamics, in
    conjunction with a numerical integration function.
    
    @type robot: Robot object, n-axes
    @param robot: The robot
    @rtype: n-vector
    @return: Joint coordinate acceleration
    @bug: Should handle the case of q, qd are matrices
    @see: L{rne}, L{robot}
    """
    n = robot.n
    if len(args) == 1:
        q = mat(args[0])[0,0:n]
        qd = mat(args[0])[0,n:2*n]
        torque = mat(args[0])[0,2*n:3*n]
    else:
        q = mat(args[0])
        if numcols(q) == robot.n:
            q = q.flatten().T
            qd = mat(args[1]).flatten().T
        torque = mat(args[2])
    
    # Compute current manipulator inertia
    # Torques resulting from unit acceleration of each joint with
    # no gravity

    M = rne(robot, mat(ones((n,1)))*q.T, zeros((n,n)), eye(n), [[0],[0],[0]])

    # Compute gravity and coriolis torque
    # torques resulting from zero acceleration at given velocity &
    # with gravity acting

    tau = rne(robot, q.T, qd.T, zeros((1,n)))
    qdd = inv(M) * (torque.flatten().T - tau.T)
    return qdd.T

def coriolis(robot, q, qd):
    """
    Compute the manipulator Coriolis matrix

    c = coriolis(robot, q, qd)

    Returns the n-element Coriolis/centripetal torque vector at the specified 
    pose and velocity.

    If C{q} and C{qd} are row vectors, the result is a row vector 
    of joint torques.
    If C{q} and C{qd} are matrices, each row is interpretted as a joint state 
    vector, and the result is a matrix each row being the corresponding joint torques.

    @type q: M{m x n} matrix
    @type q: Joint coordinate
    @type qd: M{m x n} matrix
    @type qd: Joint coordinate velocity
    @type robot: Robot object, n-axes
    @param robot: The robot
    @rtype: M{m x n} matrix
    @return: Joint coordinate acceleration
    @see: L{robot}, L{rne}, L{itorque}, L{gravload}
    """
    return rne(robot, q, qd, zeros(shape(q)), [[0],[0],[0]])

def inertia(robot, q):
    """
    Compute the manipulator inertia matrix

    inertia(robot, q)

    Returns the M{n x n} symmetric inertia matrix which relates joint torque 
    to joint acceleration for the robot in pose C{q}.
    
    If C{q} is a matrix then return a list of inertia matrices, corresponding
    to the joint coordinates from rows of C{q}.

    @type q: M{m x n} matrix
    @type q: Joint coordinate
    @type robot: Robot object, n-axes
    @param robot: The robot
    @rtype: m-list of M{n x n} matrices
    @return: List of inertia matrices
    @see: L{cinertia}, L{rne}, L{robot}
    """
    n = robot.n
    q = mat(q)
    if numrows(q) > 1:
        ilist = [];
        for row in q:
            I = rne(robot, ones((n, 1))*row, zeros((n,n)), eye(n), [[0],[0],[0]])
            ilist.append( I )
        return ilist
    else:
        return rne(robot, ones((n, 1))*q, zeros((n,n)), eye(n), [[0],[0],[0]])

def cinertia(robot, q):
    """
    Compute the Cartesian (operational space) manipulator inertia matrix

    m = cinertia(robot, q)

    Return the M{6 x 6} inertia matrix which relates Cartesian force/torque to 
    Cartesian acceleration.

    @type q: n-vector
    @type q: Joint coordinate
    @type robot: Robot object, n-axes
    @param robot: The robot
    @rtype: M{6 x 6} matrix
    @return: Cartesian inertia matrix
    @bug: Should handle the case of q as a matrix
    @see: L{inertia}, L{rne}, L{robot}
    """
    
    J = Jac.jacob0(robot, q)
    Ji = inv(J)
    M = inertia(robot, q)
    return Ji.T*M*Ji
    
def gravload(robot, q, gravity=None):
    """
    Compute the gravity loading on manipulator joints

    taug = gravload(robot, q)
    taug = gravload(robot, q, grav)

    Compute the joint gravity loading for the manipulator C{robot} in the
    configuration C{q}.

    If C{q} is a row vector, the result is a row vector of joint torques.
    If C{q} is a matrix, each row is interpretted as a joint state vector, and
    the result is a matrix each row being the corresponding joint torques.

    Gravity vector can be given explicitly using the named gravity keyword, otherwise
    it defaults to the value of the C{robot} object.

    @type q: M{m x n} matrix
    @type q: Joint coordinate
    @type grav: 3-vector
    @type grav: Gravitation acceleration, overrides C{robot} (optional)
    @type robot: Robot object, n-axes
    @param robot: The robot
    @rtype: M{m x n} matrix
    @return: Gravity torque
    @see: L{rne}, L{robot}
    """
    q = mat(q)
    if numcols(q) != robot.n:
        raise 'Insuficient columns in q'
    if gravity == None:
        tg = rne(robot, q, zeros(shape(q)), zeros(shape(q)))
    else:
        tg = rne(robot, q, zeros(shape(q)), zeros(shape(q)), gravity=gravity)
    return tg


def rne(robot, *args, **options):
    """
    Compute inverse dynamics via recursive Newton-Euler formulation.

    tau = rne(robot, q, qd, qdd)
    tau = rne(robot, [q qd qdd])

    Returns the joint torque required to achieve the specified joint position,
    velocity and acceleration state.

    Options
    =======
    
        One or more options can be provided as named arguments.
        
        Gravity
        -------
        Gravity vector is an attribute of the robot object but this may be 
        overriden by providing a gravity acceleration vector [gx gy gz] using the
        named argument gravity

        tau = rne(robot, ..., gravity=[gx, gy, gz])


        External force/moment
        ---------------------
        An external force/moment acting on the end of the manipulator may also be
        specified by a 6-element vector [Fx Fy Fz Mx My Mz].

        tau = rne(robot, ..., fext=[fx,fy,fz, mx,my,mz])

        where Q, QD and QDD are row vectors of the manipulator state; pos, vel, 
        and accel.
        
        Debug
        -----

        tau = rne(robot, ..., debug=n)
        
        Use the debug named argument to enable
            - 0 no messages
            - 1 display results of forward and backward recursions
            - 2 display print R and p*

    @see: L{robot}, L{accel}, L{inertia}, L{coriolis}, L{gravload}
    @note: verified against Matlab toolbox
    """

    n = robot.n
    
    a1 = mat(args[0])
    if numcols(a1)== 3*n:
        # single state parameter: [q qd qdd]
        q = a1[:,0:n]
        qd = a1[:,n:2*n]
        qdd = a1[:,2*n:3*n]
    else:
        # three state parameters: q, qd, qdd
        np = numrows(a1)
        q = a1
        qd = mat(args[1])
        qdd = mat(args[2])
        if numcols(a1) != n or numcols(qd) != n or numcols(qdd) != n or numrows(qd) != np or numrows(qdd) != np:
            error('inconsistant sizes of q, qd, qdd')

    # process options: gravity,  fext, debug
    debug = 0;
    gravity = robot.gravity;
    fext = mat(zeros((6,1)))
    
    for k,v in options.items():
        if k == "gravity":
            gravity = arg2array(v);
            if not isvec(gravity, 3):
                error('gravity must be 3-vector')
            gravity = mat(gravity).T
        elif k == "fext":
            fext = arg2array(v);
            if not isvec(fext, 6):
                error('fext must be 6-vector')
            fext = mat(fext).T
        elif k == "debug":
            debug = v
        else:
            error('unknown option')
        
    if robot.ismdh():
        tau = _rne_mdh(robot, q, qd, qdd, gravity, fext, debug=debug)
    else:
        tau = _rne_dh(robot, q, qd, qdd, gravity, fext, debug=debug)
    return tau



def _rne_dh(robot, Q, Qd, Qdd, grav, fext, debug=0):
  
    z0 = mat([0,0,1]).T
    n = robot.n
    np = numrows(Q)
    tau = mat(zeros((np,n)))
   
    for p in range(0,np):
        q = Q[p,:].T
        qd = Qd[p,:].T
        qdd = Qdd[p,:].T

        Fm = []
        Nm = []
        pstarm = []
        Rm = []
        w = mat(zeros((3,1)))
        wd = mat(zeros((3,1)))
        v = mat(zeros((3,1)))
        vd = grav

        #
        # init some variables, compute the link rotation matrices
        #
        for j in range(0,n):
            link = robot.links[j]
            Tj = link.tr(q[j,0])
            Rm.append(t2r(Tj))
            if link.sigma == 0:
                D = link.d
            else:
                D = q[j,0]
            alpha = link.alpha
            pstarm.append(mat([[link.a],[D*sin(alpha)],[D*cos(alpha)]]))
        
        #
        # the forward recursion
        #
        for j in range(0,n):
            link = robot.links[j]
            Rt = Rm[j].T   # transpose!!
            pstar = pstarm[j]
            r = link.r

            #
            # statement orden is important here
            #
            
            if link.sigma == 0:
                # revolute axis
                wd = Rt*(wd + z0*qdd[j,0]+crossp(w,z0*qd[j,0]))
                w = Rt*(w + z0*qd[j,0])
                #v = crossp(w,pstar) + Rt*v
                vd = crossp(wd,pstar) + crossp(w,crossp(w,pstar)) + Rt*vd
            
            else:
                # prismatic axis
                w = Rt*w
                wd = Rt*wd
                vd = Rt*(z0*qdd[j,0]+vd) + crossp(wd,pstar) + 2*crossp(w,Rt*z0*qd[j,0]) + crossp(w,crossp(w,pstar))

            vhat = crossp(wd,r) + crossp(w,crossp(w,r)) + vd
            F = link.m*vhat
            N = link.I*wd + crossp(w,link.I*w)
            Fm.append(F)
            Nm.append(N)

        #
        # the backward recursion
        #

        fext = fext.flatten().T
        f = fext[0:3,0]           # force/moments on end of arm
        nn = fext[3:6,0]
        
        for j in range(n-1,-1,-1):
            link = robot.links[j]
            pstar = pstarm[j]

            #
            # order of these statements is important, since both
            # mm and f are functions of previous f.
            #

            if j == n-1:
                R = mat(eye(3,3))
            else:
                R = Rm[j+1]
            r = link.r.T

            nn = R*(nn + crossp(R.T*pstar,f)) + crossp(pstar+r,Fm[j]) + Nm[j]
            f = R*f + Fm[j]

            R = Rm[j]
            if link.sigma == 0:
                # revolute
                tau[p,j] = nn.T*(R.T*z0) + link.G**2*link.Jm*qdd[j,0] + link.G*link.friction(qd[j,0])
            else:
                # prismatic
                tau[p,j] = f.T*(R.T*z0) + link.G**2*link.Jm*qdd[j,0] + link.G*link.friction(qd[j,0])
    return tau


def _rne_mdh(robot, Q, Qd, Qdd, grav, fext, debug=0):

    z0 = mat([0,0,1]).T
    n = robot.n
    np = numrows(Q)
    tau = mat(zeros((np,n)))
        
    for p in range(0,np):
        q = Q[p,:].T
        qd = Qd[p,:].T
        qdd = Qdd[p,:].T

        Fm = []
        Nm = []
        Pm = []
        Rm = []
        w = mat(zeros((3,1)))
        wd = mat(zeros((3,1)))
        v = mat(zeros((3,1)))
        vd = grav.flatten().T

        #
        # init some variables, compute the link rotation matrices
        #
        for j in range(0,n):
            link = robot.links[j]
            Tj = link.tr(q[j,0])
            Rm.append(t2r(Tj))
            if link.sigma == 0:
                D = link.d
            else:
                D = q[j,0]
            alpha = link.alpha
            Pm.append(mat([[link.a],[-D*sin(alpha)],[D*cos(alpha)]])) # (i-1) P i

        #
        # the forward recursion
        #
        for j in range(0,n):
            link = robot.links[j]
            R = Rm[j].T   # transpose!!
            P = Pm[j]
            Pc = link.r

            #
            # trailing underscore means new value
            #
            
            if link.sigma == 0:
                # revolute axis
                w_ = R*w + z0*qd[j,0]
                wd_ = R*wd + crossp(R*w,z0*qd[j,0]) + z0*qdd[j,0]
                #v = crossp(w,P) + R*v
                vd_ = R*(crossp(wd,P) + crossp(w,crossp(w,P)) + vd)
            
            else:
                # prismatic axis
                w_ = R*w
                wd_ = R*wd
                #v = R*(z0*qd[j,0] + v) + crossp(w,P)
                vd_ = R*(crossp(wd,P) + crossp(w,crossp(w,P)) + vd) + 2*crossp(R*w,z0*qd[j,0]) + z0*qdd[j,0]

            # update variables
            w = w_
            wd = wd_
            vd = vd_

            vdC = crossp(wd,Pc) + crossp(w,crossp(w,Pc)) + vd
            F = link.m*vdC
            N = link.I*wd + crossp(w,link.I*w)
            Fm.append(F)
            Nm.append(N)


        #
        # The backward recursion
        #

        fext = fext.flatten().T
        f = fext[0:3,0]           # force/moments on end of arm
        nn = fext[3:6,0]

        for j in range(n-1,-1,-1):
            link = robot.links[j]

            #
            # order of these statements is important, since both
            # mm and f are functions of previous f.
            #

            if j == n-1:
                R = mat(eye(3,3))
                P = mat([[0],[0],[0]])
            else:
                R = Rm[j+1]
                P = Pm[j+1]   # i/P/(i+1)
            Pc = link.r
            f_ = R*f + Fm[j]
            nn_ = Nm[j] + R*nn + crossp(Pc,Fm[j]) + crossp(P,R*f)
            f = f_
            nn = nn_
            
            if link.sigma == 0:
                # revolute
                tau[p,j] = nn.T*z0 + link.G**2*link.Jm*qdd[j,0] + link.G*link.friction(qd[j,0])
            else:
                # prismatic
                tau[p,j] = f.T*z0 + link.G**2*link.Jm*qdd[j,0] + link.G*link.friction(qd[j,0])
    return tau




'''Utilities to reconstruct BMI trajectories'''
import numpy as np
import os
import tables
import pickle

#import bmimultitasks, cursor_clda_tasks
from riglib.bmi.bmi import BMILoop
from riglib.bmi import extractor, clda, feedback_controllers, goal_calculators
from riglib.experiment import Experiment, traits
from features.simulation_features import SimHDF
from riglib import plants

def sim_target_seq_generator_multi(n_targs, n_trials):
    '''
    Simulated generator for simulations of the BMIControlMulti and CLDAControlMulti tasks
    '''
    center = np.zeros(2)
    pi = np.pi
    targets = 8*np.vstack([[np.cos(pi/4*k), np.sin(pi/4*k)] for k in range(8)])

    target_inds = np.random.randint(0, n_targs, n_trials)
    target_inds[0:n_targs] = np.arange(min(n_targs, n_trials))
    for k in range(n_trials):
        targ = targets[target_inds[k], :]
        yield np.array([[center[0], 0, center[1]],
                        [targ[0], 0, targ[1]]])


class BMIReconstruction(BMILoop, Experiment):
    fps = 60
    def __init__(self, n_iter=None, entry_id=None, hdf_filename=None, decoder_filename=None, params=dict(), *args, **kwargs):
        if entry_id is None and (hdf_filename is None or decoder_filename is None):
            raise ValueError("Not enough data to reconstruct a BMI! Specify a database entry OR an HDF file + decoder file")
        if entry_id is not None:
            from db import dbfunctions as dbfn
            te = dbfn.TaskEntry(entry_id)
            self.hdf_ref = te.hdf
            self.decoder = te.decoder
            self.params = te.params

            if self.hdf_ref is None:
                raise ValueError("Database is unable to locate HDF file!")
            if self.decoder is None:
                raise ValueError("Database is unable to locate HDF file!")
        elif hdf_filename is not None:
            self.hdf_ref = tables.open_file(hdf_filename)
            self.decoder = pickle.load(open(decoder_filename, 'rb'), encoding='latin1') # extra args to get py3 to read py2 pickles
            self.params = params


        self.n_iter = min(n_iter, len(self.hdf_ref.root.task))

        try:
            self.starting_pos = self.hdf_ref.root.task[0]['decoder_state'][0:3,0]
        except:
            # The statement above appears to not always work...
            self.starting_pos = self.hdf_ref.root.task[0]['cursor'] # #(0, 0, 0)

        # if 'plant_type' in te.params:
        #     self.plant_type = te.params['plant_type']
        # elif 'arm_class' in te.params:
        #     plant_type = te.params['arm_class']
        #     if plant_type == 'CursorPlant':
        #         self.plant_type = 'cursor_14x14'
        #     else:
        #         self.plant_type = plant_type
        # else:
        #     self.plant_type = 'cursor_14x14'

        # TODO overly specific
        self.plant = plants.CursorPlant(endpt_bounds=(-14, 14, 0., 0., -14, 14))

        ## Set the target radius because the old assist method changes the assist speed
        # when the cursor is inside the target
        self.target_radius = params['target_radius']
        self.cursor_radius = params['cursor_radius']
        self.assist_level = params['assist_level']

        self.idx = 0
        gen = sim_target_seq_generator_multi(8, 1000)
        super(BMIReconstruction, self).__init__(gen, *args, **kwargs)
        self.hdf = SimHDF()
        self.learn_flag = True

        task_msgs = self.hdf_ref.root.task_msgs[:]
        self.update_bmi_msgs = task_msgs[task_msgs['msg'] == 'update_bmi']
        task_msgs = list(filter(lambda x: x['msg'] not in ['update_bmi'], task_msgs))
        # print task_msgs
        self.task_state = np.array([None]*n_iter)
        for msg, next_msg in zip(task_msgs[:-1], task_msgs[1:]):
            self.task_state[msg['time']:next_msg['time']] = msg['msg']

        self.update_bmi_inds = np.zeros(len(self.hdf_ref.root.task))
        self.update_bmi_inds[self.update_bmi_msgs['time']] = 1
        self.recon_update_bmi_inds = np.zeros(len(self.hdf_ref.root.task))

        self.target_hold_msgs = list(filter(lambda x: x['msg'] in ['target', 'hold'], self.hdf_ref.root.task_msgs[:]))

    def calc_recon_error(self, n_iter_betw_fb=100000, **kwargs):
        """Main function to call for reconstruction"""
        saved_state = self.hdf_ref.root.task[:]['decoder_state']
        while self.idx < self.n_iter:
            # print self.current_assist_level
            self.get_cursor_location(**kwargs)

            if self.idx % n_iter_betw_fb == 0:
                if saved_state.dtype == np.float32:
                    error = saved_state[:self.idx,:,-1] - np.float32(self.decoder_state[:self.idx,:,-1])
                else:
                    error = saved_state[:self.idx,:,-1] - self.decoder_state[:self.idx,:,-1]
                print("Error after %d iterations" % self.idx, np.max(np.abs(error)))

        if saved_state.dtype == np.float32:
            error = saved_state[:self.n_iter,:,-1] - np.float32(self.decoder_state[:self.n_iter,:,-1])
        else:
            error = saved_state[:self.n_iter,:,-1] - self.decoder_state[:self.n_iter,:,-1]

        return error

    def init_decoder_state(self):
        '''
        Initialize the state of the decoder to match the initial state of the plant
        '''
        init_decoder_state = self.hdf_ref.root.task[0]['decoder_state']
        if init_decoder_state.shape[1] > 1:
            init_decoder_state = init_decoder_state[:,0].reshape(-1,1)

        self.init_decoder_mean = init_decoder_state
        self.decoder.filt.state.mean = self.init_decoder_mean

        self.decoder.set_call_rate(self.fps)

    def create_feature_extractor(self):
        '''
        Create the feature extractor object
        '''
        self.extractor = extractor.ReplaySpikeCountsExtractor(self.hdf_ref.root.task,
            source='spike_counts', units=self.decoder.units, cycle_rate=self.fps)
        self._add_feature_extractor_dtype()

    def load_decoder(self):
        '''
        Create the object for the initial decoder
        '''
        self.n_subbins = self.decoder.n_subbins
        self.decoder_state = np.zeros([self.n_iter, self.decoder.n_states, self.n_subbins])

    def get_spike_counts(self):
        return self.hdf_ref.root.task[self.idx]['spike_counts']

    def _update_target_loc(self):
        self.target_location = self.hdf_ref.root.task[self.idx]['target']
        self.state = self.task_state[self.idx]

    def get_cursor_location(self, verbose=False):
        if self.idx % 1000 == 0 and verbose: print(self.idx)

        self.current_assist_level = self.hdf_ref.root.task[self.idx]['assist_level'][0]
        try:
            self.current_half_life = self.hdf_ref.root.task[self.idx]['half_life'][0]
        except:
            self.current_half_life = 0
        self._update_target_loc()

        self.call_decoder_output = self.move_plant(half_life=self.current_half_life)
        if verbose:
            print(self.call_decoder_output)
            print(self.hdf_ref.root.task[self.idx]['decoder_state'])
            print()
        self.decoder_state[self.idx] = self.call_decoder_output
        self.idx += 1

    def call_decoder(self, neural_obs, target_state, **kwargs):
        '''
        Run the decoder computations

        Parameters
        ----------
        neural_obs : np.array of shape (n_features, n_subbins)
            n_features is the number of neural features the decoder is expecting to decode from.
            n_subbins is the number of simultaneous observations which will be decoded (typically 1)
        target_state: np.array of shape (n_states, 1)
            The current optimal state to be in to accomplish the task. In this function call, this gets
            used when adapting the decoder using CLDA
        '''
        # Get the decoder output
        decoder_output, update_flag = self.bmi_system(neural_obs, target_state, self.state, learn_flag=self.learn_flag, **kwargs)
        if update_flag:
            # send msg to hdf file to indicate decoder update
            self.hdf_ref.sendMsg("update_bmi")
            self.recon_update_bmi_inds[self.idx] = 1

        return decoder_output

    def get_time(self):
        t = self.idx * 1./self.fps
        return t





'''
For capturing frames from a camera
'''
import time
import os
import traceback
from riglib.experiment import traits
from riglib.mp_calc import MultiprocShellCommand
from datetime import datetime
from riglib.e3vision import E3VisionInterface

class E3Video(traits.HasTraits):
    '''
    Enables recording of e3vision cameras from White-Matter.
    '''

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        '''
        super().init()

        # Start the natnet client and recording
        now = datetime.now()
        take = now.strftime("Take %Y-%m-%d %H_%M_%S")
        self.e3v_status = 'offline'
        try:
            if self.saveid is not None:
                take += " (%d)" % self.saveid
                self.e3v_filename = take
                e3v = E3VisionInterface(take)
                e3v.update_camera_status()
                e3v.start_rec()
                self.e3v_status = 'recording'
            else:
                e3v = E3VisionInterface()
                e3v.update_camera_status()
                self.e3v_status = 'online'
            self.e3v = e3v
        except Exception as e:
            self.e3v_status = traceback.format_exc()

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run().
        '''
        if not self.e3v_status in ['recording', 'online']:
            import io
            self.terminated_in_error = True
            self.termination_err = io.StringIO()
            self.termination_err.write(self.e3v_status)
            self.termination_err.seek(0)
            self.state = None
            super().run()
            try:
                self.e3v.stop_rec()
            except:
                pass
        else:
            try:
                super().run()
            finally:
                if self.e3v_status == 'recording':
                    print("Stopping e3v recording")
                    self.e3v.stop_rec()

    def cleanup(self, database, saveid, **kwargs):
        '''
        Save the e3v recorded filenames into the database. Only gets called when saveid is not None.
        '''
        super_result = super().cleanup(database, saveid, **kwargs)
        print("Saving WM e3vision files to database...")
        try:
            suffix = '' # note: database functions don't take keyword arguements like custom_suffix=suffix
            database.save_data(self.e3v_filename, "e3v", saveid, False, False, suffix) # Make sure you actually have an "e3v" system added!
        except Exception as e:
            print(e)
            return False
        print("...done.")
        return super_result
    

class SingleChannelVideo(traits.HasTraits):
    def __init__(self, *args, **kwargs):
        super(SingleChannelVideo, self).__init__(*args, **kwargs)

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' spawns process to run ssh command to begin video recording
        '''
        self.video_basename = 'video_%s.avi' % time.strftime('%Y_%m_%d_%H_%M_%S')        
        self.device_name = '/dev/video0'
        cmd = "ssh -tt video /home/lab/bin/recording_start.sh %s %s" % (self.device_name, self.video_basename)
        
        cmd_caller = MultiprocShellCommand(cmd)
        cmd_caller.start()
        self.cmd_caller = cmd_caller
        print("started video recording")
        print(cmd)
        
        super(SingleChannelVideo, self).init()

    def cleanup(self, database, saveid, **kwargs):
        '''
        Stop video recording and link file to database

        Parameters
        ----------
        database
        saveid

        Returns
        -------
        None
        '''

        print("executing command to stop video recording")
        os.popen('ssh -tt video /home/lab/bin/recording_end.sh')

        print("Checking if video recording is done")
        if self.cmd_caller.done.is_set():
            print("SSH command finished!")

        super(SingleChannelVideo, self).cleanup(database, saveid, **kwargs)
        print("sleeping so that video recording can wrap up")
        time.sleep(5)        

        ## Get the video filename
        video_fname = os.path.join('/storage/video/%s' % self.video_basename)

        dbname = kwargs['dbname'] if 'dbname' in kwargs else 'default'
        if dbname == 'default':
            database.save_data(video_fname, "video", saveid)
        else:
            database.save_data(video_fname, "video", saveid, dbname=dbname)



import time
import os
import numpy as np
from OpenGL.GL import *
import pygame
try:
    import xr
except:
    print("No OpenXR support")
    
from ..experiment import traits
from .render import stereo, render, ssao, shadow_map
from .models import Group
from .xfm import Quaternion, Transform
from .window import Window
from .environment import Grid

class Clock():

    def __init__(self, n_ticks=10):
        self.start_time = self.get_time()
        self.prev_ticks = np.zeros((n_ticks,))

    def tick(self, fps):
        self.prev_ticks = np.roll(self.prev_ticks, 1)
        self.prev_ticks[0] = self.get_time()

    def get_time(self):
        return time.perf_counter()
    
    def get_fps(self):
        return -1/np.mean(np.diff(self.prev_ticks))
    
class WindowVR(Window):
    '''
    An OpenXR window for rendering in VR to an HMD
    '''
    
    show_grid = traits.Bool(True, desc="Show a textured grid on the floor")
    grid_size = traits.Float(130, desc="Size of the grid in cm")
    grid_position = traits.Tuple((0, 0, 0), desc="Position of the grid in cm. If you want the floor of the grid to be on the floor of the world, set the z component to (grid_size - camera_offset[2])")
    camera_offset = traits.Tuple((0, -130, 40), desc="Offset virtual screen to the camera in cm")
    camera_position = traits.Tuple((0, 0, -40), desc="Absolute position of the camera (x,y,z) in cm world coordinates. Only used if fixed_camera_position is True")
    camera_orientation = traits.Tuple((1, 0, 0, 0), desc="Orientation of the camera (w, x, y, z) as a quaternion. Only used if fixed_camera_orientation is True")
    fixed_camera_position = traits.Bool(False, desc="Fixed position of the camera")
    fixed_camera_orientation = traits.Bool(False, desc="Fixed orientation of the camera")

    hidden_traits = ['fps', 'window_size', 'screen_dist']

    def init(self):
        self.add_dtype('view_pose_position', 'f8', (2,3))
        self.add_dtype('view_pose_rotation', 'f8', (2,4))
        self.add_dtype('modelview', 'f8', (2,4,4))
        super().init()

    def screen_init(self):
        from ctypes import byref, c_int32, c_void_p, cast, POINTER, pointer, Structure

        # os.environ['XR_RUNTIME_JSON'] = '/usr/share/openxr/1/openxr_monado.json'
        os.environ['XR_RUNTIME_JSON'] = '/home/aolab/.config/openxr/1/active_runtime.json'
        pygame.init()
        self.clock = Clock()
        self.fps = 90

        context = xr.ContextObject(
            instance_create_info=xr.InstanceCreateInfo(
                enabled_extension_names=[
                    xr.KHR_OPENGL_ENABLE_EXTENSION_NAME,
                ],
            ),
            reference_space_create_info=xr.ReferenceSpaceCreateInfo(
                reference_space_type=xr.ReferenceSpaceType.STAGE,
                pose_in_reference_space=xr.Posef((0,0,0,1), (0,0,0)),
            ),
            
        )
        # context.__enter__()

        '''
        Ideally we would use the context manager here, but it uses the default
        swapchain image format, which is not guaranteed to be an SRGB format.
        There might be a better way to handle this but for now this works.
        '''
        context.instance = xr.create_instance(
            create_info=context._instance_create_info,
        )
        context.system_id = xr.get_system(
            instance=context.instance,
            get_info=xr.SystemGetInfo(
                form_factor=context.form_factor,
            ),
        )

        if context._session_create_info.next is None:
            context.graphics = xr.OpenGLGraphics(
                instance=context.instance,
                system=context.system_id,
                title=context._instance_create_info.application_info.application_name.decode()
            )
            context.graphics_binding_pointer = cast(pointer(context.graphics.graphics_binding), c_void_p)
            context._session_create_info.next = context.graphics_binding_pointer
        else:
            context.graphics_binding_pointer = context._session_create_info.next

        context._session_create_info.system_id = context.system_id
        context.session = xr.create_session(
            instance=context.instance,
            create_info=context._session_create_info,
        )
        context.space = xr.create_reference_space(
            session=context.session,
            create_info=context._reference_space_create_info
        )
        context.default_action_set = xr.create_action_set(
            instance=context.instance,
            create_info=xr.ActionSetCreateInfo(
                action_set_name="default_action_set",
                localized_action_set_name="Default Action Set",
                priority=0,
            ),
        )
        context.action_sets.append(context.default_action_set)

        # Create swapchains
        config_views = xr.enumerate_view_configuration_views(
            instance=context.instance,
            system_id=context.system_id,
            view_configuration_type=context.view_configuration_type,
        )
        context.graphics.initialize_resources()
        swapchain_formats = xr.enumerate_swapchain_formats(context.session)
        color_swapchain_format = context.graphics.select_color_swapchain_format(swapchain_formats) # Ignore this
        # Create a swapchain for each view.
        context.swapchains.clear()
        context.swapchain_image_buffers.clear()
        context.swapchain_image_ptr_buffers.clear()
        for vp in config_views:
            # Create the swapchain.
            swapchain_create_info = xr.SwapchainCreateInfo(
                array_size=1,
                format=GL_SRGB8_ALPHA8, # Set to SRGB format otherwise the colors are washed out
                width=vp.recommended_image_rect_width,
                height=vp.recommended_image_rect_height,
                mip_count=1,
                face_count=1,
                sample_count=vp.recommended_swapchain_sample_count,
                usage_flags=xr.SwapchainUsageFlags.SAMPLED_BIT | xr.SwapchainUsageFlags.COLOR_ATTACHMENT_BIT,
            )
            swapchain = xr.context_object.SwapchainStruct(
                xr.create_swapchain(
                    session=context.session,
                    create_info=swapchain_create_info,
                ),
                swapchain_create_info.width,
                swapchain_create_info.height,
            )
            context.swapchains.append(swapchain)
            swapchain_image_buffer = xr.enumerate_swapchain_images(
                swapchain=swapchain.handle,
                element_type=context.graphics.swapchain_image_type,
            )
            # Keep the buffer alive by moving it into the list of buffers.
            context.swapchain_image_buffers.append(swapchain_image_buffer)
            capacity = len(swapchain_image_buffer)
            swapchain_image_ptr_buffer = (POINTER(xr.SwapchainImageBaseHeader) * capacity)()
            for ix in range(capacity):
                swapchain_image_ptr_buffer[ix] = cast(
                    byref(swapchain_image_buffer[ix]),
                    POINTER(xr.SwapchainImageBaseHeader))
            context.swapchain_image_ptr_buffers.append(swapchain_image_ptr_buffer)
        context.graphics.make_current()
        '''
        End of context initialization
        '''

        # Query the swapchain size
        config_views = xr.enumerate_view_configuration_views(
            instance=context.instance,
            system_id=context.system_id,
            view_configuration_type=context.view_configuration_type,
        )
        self.window_size = (
            config_views[0].recommended_image_rect_width * 2,
            config_views[0].recommended_image_rect_height)

        glDisable(GL_FRAMEBUFFER_SRGB)
        glEnable(GL_BLEND)
        glDepthFunc(GL_LESS)
        glEnable(GL_DEPTH_TEST)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glClearColor(*self.background)
        glClearDepth(1.0)
        glDepthMask(GL_TRUE)
        glEnable(GL_CULL_FACE)
        glCullFace(GL_BACK)

        self.renderer = self._get_renderer()

        if self.show_grid:
            self.add_model(Grid(self.grid_size*2).translate(self.grid_position[0], self.grid_position[1], self.grid_position[2]))
        self.world = Group(self.models)
        self.world.init()
        self.set_eye((0,0,0), (0,0))
        self.xr_frame_generator = context.frame_loop()
        self.xr_context = context
        print("Initialized OpenXR window")

    def _get_renderer(self):
        near = 1
        far = 1024
        if self.stereo_mode == 'mirror':
            glFrontFace(GL_CW);  # Switch to clockwise winding for mirrored objects
        return shadow_map.ShadowMapper(self.window_size, self.fov, near, far)
    
    def draw_world(self):
        # Get the OpenXR views
        try:
            frame_state = next(self.xr_frame_generator)
        except StopIteration:
            self.state = None  # Exit loop if the generator is exhausted

        for view_index, view in enumerate(self.xr_context.view_loop(frame_state)):
            projection = xr.Matrix4x4f.create_projection_fov(
                graphics_api=xr.GraphicsAPI.OPENGL,
                fov=view.fov,
                near_z=0.05,
                far_z=1024,
            ).as_numpy().reshape(4,4).T
            if self.fixed_camera_position:
                position = self.camera_position - np.array([1,0,0])*self.iod*(view_index-0.5)
            else:
                position = -np.array([
                    view.pose.position[0]*100 + self.camera_offset[0],
                    view.pose.position[1]*100 + self.camera_offset[1],
                    view.pose.position[2]*100 + self.camera_offset[2],
                ]) # Not sure why this needs to be negated, something to do with the handedness of the coordinate system??
            if self.fixed_camera_orientation:
                rotation = self.camera_orientation
            else:
                rotation = np.array([
                    -view.pose.orientation.w, # Also not sure why I need to negate the w component
                    view.pose.orientation.x,
                    view.pose.orientation.y,
                    view.pose.orientation.z,
                ])
            xfm = Transform(move=position, rotate=Quaternion(*rotation)) 
            self.modelview = xfm.to_mat(reverse=True)

            # Optionally mirror the view along the y-axis
            if self.stereo_mode == 'mirror':
                self.modelview = np.dot(self.modelview, np.diag([-1,1,1,1]))

            # Draw the world
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
            self.renderer.draw(self.world, p_matrix=projection, modelview=self.modelview)
        
            # Save the pose data
            if hasattr(self, 'task_data'):
                self.task_data['view_pose_position'][:,view_index,:] = position
                self.task_data['view_pose_rotation'][:,view_index,:] = rotation
                self.task_data['modelview'][:,view_index] = self.modelview

        self.renderer.draw_done()

    def _test_stop(self, ts):
        super_stop = super(Window, self)._test_stop(ts)
        return super_stop

    def _start_None(self):
        self.xr_context.__exit__(None, None, None)
        super(WindowVR, self)._start_None()



import serial
import pyfirmata, time
import serial.tools.list_ports
import numpy as np
from riglib.lasers import qwalor_laser
from riglib.lasers import optical_switch

# The following command can be used for finding available ports
# python -m serial.tools.list_ports

import unittest

# laser setting
channel = 2
mode = 'off'
freq = 0

# pulse_width_list = [0.00001, 0.00002, 0.00003, 0.00005, 0.0001, 0.0002, 0.0003, 0.0005, 0.001, 0.002, 0.003, 0.005, 0.01, 0.02, 0.03, 0.05, 0.1, 0.2, 0.3, 0.5, 1]
pulse_width_list = [1]
power_list = [0.5, 0.6, 0.7, 0.8, 0.9, 1]

class LaserTests(unittest.TestCase):

    @unittest.skip("only run if needed")
    def test_config(self):
        packet = qwalor_laser.get_config_packet(1, 0, 0, debug=True)
        self.assertEqual(np.count_nonzero(packet), 0)

        packet = qwalor_laser.get_config_packet(2, 0.5137, 0.5, mode='SW', debug=True)
        self.assertEqual(np.count_nonzero(packet), 4)
        self.assertEqual(packet[1], 1)
        self.assertEqual(packet[2], 0xff)
        self.assertEqual(packet[3], 0x7f)

    def test_laser_setup(self):
        config_packet = qwalor_laser.get_config_packet(channel, freq, 1., mode)
        print(config_packet)
        link = qwalor_laser.QwalorLink.get_instance()
        link.send(config_packet)

    @unittest.skip("only run if needed")
    def test_run(self):

        laser = qwalor_laser.QwalorLaserSerial(channel)
        laser.set_mode(mode)
        laser.set_freq(freq)

        for power in power_list:
            laser.set_power(power)
            print(power)
            for width in pulse_width_list:
                t0 = time.perf_counter()
                laser.on()
                while (time.perf_counter() - t0 < width):
                    pass
                laser.off()
                time.sleep(0.5)

    @unittest.skip("only run if needed")
    def test_dual_channel(self):
        n_trials = 10
        iti = 0.1
        width = 0.1

        laser1 = qwalor_laser.QwalorLaserSerial(2)
        laser1.port = laser1.trigger_pin
        laser1.name = 'qwalor_laser_ch2'
        print("laser 1 configured")

        laser2 = qwalor_laser.QwalorLaserSerial(4)
        laser2.port = laser2.trigger_pin
        laser2.name = 'qwalor_laser_ch4'
        print("laser 2 configured")

        for n in range(n_trials):
            t0 = time.perf_counter()
            laser1.set_mode(mode)
            laser1.set_freq(freq)
            laser1.set_power(1)
            laser1.on()
            while (time.perf_counter() - t0 < width):
                pass
            laser1.off()
            t1 = time.perf_counter()
            while (time.perf_counter() - t1 < iti):
                pass
        
            t0 = time.perf_counter()
            laser2.set_mode(mode)
            laser2.set_freq(freq)
            laser2.set_power(1)
            laser2.on()
            while (time.perf_counter() - t0 < width):
                pass
            laser2.off()
            t1 = time.perf_counter()
            while (time.perf_counter() - t1 < iti):
                pass


    @unittest.skip("only run if needed")
    def test_speed(self):

        laser = qwalor_laser.QwalorLaserSerial(channel)
        laser.set_mode(mode)
        laser.set_freq(freq)

        n_trials = 1000
        iti = 0.001
        width = 0.001

        t_begin = time.perf_counter()

        for n in range(n_trials):
            laser.set_power(1)
            t0 = time.perf_counter()
            laser.on()
            while (time.perf_counter() - t0 < width):
                pass
            laser.off()
            t1 = time.perf_counter()
            while (time.perf_counter() - t1 < iti):
                pass

        t_end = time.perf_counter()
        print(f"laser took {t_end-t_begin:.2f} s to run {n_trials} trials of {width} width and {iti} iti")
        time.sleep(0.5)
        laser.off()

class OpticalSwitchTests(unittest.TestCase):

    @unittest.skip("only run if needed")
    def test_connect(self):

        switch = serial.Serial('/dev/ttyACM1', 115200, timeout=10)
        
        # Check status
        switch.write(bytearray([17]))
        ret = int.from_bytes(switch.read(), 'big')
        print()
        print(f"ready: {ret}")

    def test_class(self):
        switch = optical_switch.LFiberOpticalSwitch('/dev/ttyACM1')
        ret = switch.check_status()
        print(f'class returned: {ret}')

        t0 = time.perf_counter()
        ret = switch.set_channel(1)
        t1 = time.perf_counter()
        print(f"switch took {1000*(t1 - t0):0.2f} ms")

        time.sleep(0.1)

        t0 = time.perf_counter()
        ret = switch.reset()
        t1 = time.perf_counter()
        print(f"switch took {1000*(t1 - t0):0.2f} ms")

        time.sleep(0.1)

        t0 = time.perf_counter()
        ret = switch.set_channel(2)
        t1 = time.perf_counter()
        print(f"switch took {1000*(t1 - t0):0.2f} ms")

        time.sleep(0.1)

        t0 = time.perf_counter()
        ret = switch.set_channel(4)
        t1 = time.perf_counter()
        print(f"switch took {1000*(t1 - t0):0.2f} ms")

        time.sleep(0.1)

        t0 = time.perf_counter()
        ret = switch.set_channel(8)
        t1 = time.perf_counter()
        print(f"switch took {1000*(t1 - t0):0.2f} ms")

        time.sleep(0.1)

        t0 = time.perf_counter()
        ret = switch.set_channel(15)
        t1 = time.perf_counter()
        print(f"switch took {1000*(t1 - t0):0.2f} ms")



if __name__ == '__main__':
    unittest.main()


'''
Classes for general purpose input output (GPIO)
'''

import types
import numpy as np
#from multiprocessing import Process, Lock
from threading import Thread, Lock, Event
import time
import pyfirmata
from serial.serialutil import SerialException
import serial

from .source import DataSourceSystem

def convert_masked_data_to_pins(mask, data, bits=64):
    ''' Helper to take a mask and some data and turn it into a list of pins and values'''
    pins = []
    values = []
    for pin in range(bits):
        pin_mask = 1 << pin
        if pin_mask & mask:
            pins.append(pin)
            values.append(pin_mask & data > 0)
    return pins, values

class GPIO(object):
    ''' Wrapper for digital i/o'''

    def write(self, pin: int, value: bool):
        pass

    def read(self, pin: int) -> bool:
        pass

    def write_many(self, mask: int, data: int):
        '''
        Write data
        
        mask: sets the address to write
        data: value to write
        '''
        pass

    def read_many(self, mask: int) -> int:
        pass

class TestGPIO(GPIO):

    def __init__(self, pins=14):
        self.value = np.zeros((pins,100))

    def write(self, pin, value):
        self.value[pin,:-1,] = self.value[pin, 1:]
        self.value[pin,-1] = value
        print(".", end="")

    def write_many(self, mask, data):
        pins, values = convert_masked_data_to_pins(mask, data)
        for idx in range(len(pins)):
            self.write(pins[idx], values[idx])

    def read(self, pin):
        return self.value[pin,-1]

class CustomBoard(pyfirmata.Board):

    default_layout = pyfirmata.boards.BOARDS['arduino']
    wait_time = 1

    def __init__(self, port, layout=None, baudrate=112500, name=None, timeout=None):
        self.sp = serial.Serial(port, baudrate, timeout=timeout)
        self.pass_time(self.wait_time)
        self.name = name
        self._layout = layout
        if not self.name:
            self.name = port

        if layout:
            self.setup_layout(layout)
        else:
            layout = self.default_layout
            self.setup_layout(layout)

        # Iterate over the first messages to get firmware data
        while self.bytes_available():
            self.iterate()

class ArduinoGPIO(GPIO):
    ''' Pin-addressable arduino serial interface'''
    def __init__(self, port=None, baudrate=57600, timeout=10, enable_analog=False):
        if port is None:
            import serial.tools.list_ports
            ports = serial.tools.list_ports.comports()
            for p in ports:
                if 'USB' in p.description or 'ACM' in p.description:
                    port = p.device
            if port is None:
                raise Exception('No serial device found')
        self.board = CustomBoard(port, baudrate=baudrate, timeout=timeout)
        self.lock = Lock()
        if enable_analog:
            it = pyfirmata.util.Iterator(self.board)
            it.start()
        self.enable_analog = enable_analog

    def write(self, pin, value):
        with self.lock:
            self.board.digital[pin].write(int(value))
    
    def read(self, pin):
        with self.lock:
            return bool(self.board.digital[pin].read())

    def analog_read(self, pin):
        if not self.enable_analog:
            raise ValueError("Analog reporting not enabled, start over with enable_analog=True!")
        with self.lock:
            if self.board.analog[pin].reporting == False:
                self.board.analog[pin].enable_reporting() # This analog_read() call will fail
            return self.board.analog[pin].read()
        
    def write_many(self, mask, data):
        pins, values = convert_masked_data_to_pins(mask, data)
        for idx in range(len(pins)):
            self.board.digital[pins[idx]].write(values[idx])
        # for p in self.board.digital_ports:
        #     p.write()

    def close(self):
        ''' Call this method before destroying the object'''
        self.board.exit()

class TeensyBoard(CustomBoard):

    default_layout = {
        'digital': tuple(x for x in range(63)),
        'pwm': tuple(x for x in range(63)),
        'analog': tuple(x for x in range(43, 69)),
        'disabled': (0,1),
    }
    wait_time = 0.5 # Teensy is much faster than arduino with hardware Serial port, so we 
                    # don't have to wait so long.


class TeensyGPIO(ArduinoGPIO):

    def __init__(self, port=None, baudrate=112500, timeout=10):
        if port is None:
            import serial.tools.list_ports
            ports = serial.tools.list_ports.comports()
            for p in ports:
                if 'USB Serial' in p.description:
                    port = p.device
            if port is None:
                raise Exception('No serial device found')
        self.board = TeensyBoard(port, baudrate=baudrate, timeout=timeout)
        self.lock = Lock()

    def analog_write(self, pin, value):
        '''
        Write to any pin as a PWM (for digital pins) or analog output (if it supports DAC).
        Can fail if no connection to the board.
        '''
        with self.lock:
            sysex_cmd = 0x6F # extended analog
            data = bytearray([pin, value % 128, value >> 7]) # [pin, lsb, msb]
            try:
                self.board.send_sysex(sysex_cmd, data)
            except ValueError:
                # This happens in BCI experiment sometimes, not sure why. Ignore ~LRS
                pass 
            except SerialException:
                # This happens if the board becomes unplugged, but crashes the game if uncaught.
                raise OSError("Teensy board has become unplugged!")

class NIGPIO(GPIO):

    def __init__(self):
        from riglib.dio.NIUSB6501.py_comedi_control import init_comedi, write_to_comedi
        self.lock = Lock()
        init_comedi()
        self._write_to_comedi = write_to_comedi

    def write(self, pin, value):
        self.write_many(1, value, pin)

    def write_many(self, mask, data, base_channel=0):
        with self.lock:
            ret = self._write_to_comedi(data, mask=mask, base_channel=base_channel)
            if ret == b'':
                raise IOError("Unable to send NIDAQ sync event")

class DigitalWave(Thread):
    ''' Logic-level waves generated by a single or multiple GPIO pins'''

    def __init__(self, gpio, mask=1<<8, data=None):
        ''' Can input a single gpio or multiple'''
        self.gpio = gpio
        self.mask = mask
        if data is None:
            self.data = mask
        else:
            self.data = data
        self.edges = np.zeros(0)
        self.first_edge = True # Note, this doesn't affect the previous laser state
        self._stop_event = Event()
        Thread.__init__(self)
    
    def set_edges(self, edges, first_edge):
        ''' Directly set the list of edges'''
        self.edges = edges
        self.first_edge = first_edge

    def set_pulse(self, duration, first_edge):
        ''' Set the wave to a pulse'''
        self.set_edges([0, duration], first_edge)

    def set_square_wave(self, freq, duration, duty_cycle=0.5, phase_delay=0., first_edge=True):
        ''' Set the wave to a square wave'''
        edges = DigitalWave.square_wave(freq, duration, duty_cycle, phase_delay)
        self.set_edges(edges, first_edge)

    def run(self):
        ''' Generate square pulses defined by edges (in seconds)'''
        t0 = time.perf_counter()
        state = self.first_edge 
        for edge in self.edges:
            while (time.perf_counter() - t0 < edge):
                time.sleep(0)
                if self._stop_event.is_set(): # cancel the wave
                    if state is True:
                        return
                    edge = 0
                    self.edges = []
                    state = False
                    break
            self.gpio.write_many(self.mask, int(state*self.data))
            state = not state

    def stop(self):
        self._stop_event.set()

    @staticmethod
    def delays_to_edges(delays):
        edges = np.zeros(len(delays))
        t = 0
        for i in range(0,len(delays)):
            t += delays[i]
            edges[i] = t
        return edges

    @staticmethod
    def square_wave(freq, duration, duty_cycle=0.5, phase_delay=0):
        '''
        Generate edges for a square wave

        Arguments
        ---------
        freq - frequency of the wave
        duration - total length of the wave, in seconds
        duty_cycle - fraction of each cycle which should be in the ON state
        phase_delay - time in seconds before the start of the wave

        Returns
        -------
        edges - list of edge transition times, in seconds
        '''
        pulse_interval = 1.0/freq
        edge_interval = [pulse_interval*duty_cycle, pulse_interval*(1-duty_cycle)]
        n_edges = int(duration/pulse_interval)
        on_edges = edge_interval[0]*np.ones(n_edges)
        off_edges = edge_interval[1]*np.ones(n_edges)
        delays = np.insert(np.ravel([on_edges, off_edges], order='F'), 0, phase_delay)
        # if len(delays) % 2 == 1:
        #     delays = np.append(delays, edge_interval[1]) # always have an even number of edges
        return DigitalWave.delays_to_edges(delays)

class DigitalWaveMulti(DigitalWave):

    def __init__(self, gpio_list, mask, data):
        ''' Multiple GPIO or lasers'''
        self.edges = np.zeros(0)
        self.first_edge = True # Note, this doesn't affect the previous laser state
        self.gpio = gpio_list
        self.mask = mask
        self.data = data
        Process.__init__(self)
    
    def run(self):
        ''' Generate square pulses defined by edges (in seconds)'''
        t0 = time.perf_counter()
        state = self.first_edge 
        for edge in self.edges:
            while (time.perf_counter() - t0 < edge):
                time.sleep(0)
            for idx in range(len(self.gpio)):
                self.gpio[idx].write_many(self.mask, int(state*self.data))
            state = not state



'''
Example blank tasks for creating new tasks
'''
import numpy as np

from riglib.experiment import traits, Sequence
from riglib.stereo_opengl.window import Window
from .target_graphics import *



class ExampleSequenceTask(Window, Sequence):
    """
    Show one target after another
    """

    status = dict(
        wait = dict(start_trial="target", start_pause="pause"),
        target = dict(target_done="wait", start_pause="pause", end_state=True), # end_state is just for counting trials
        pause = dict(end_pause="wait", end_state=True),
    )

    # initial state
    state = "wait"

    sequence_generators = ['example_generator']

    example_param = traits.Float(0.0, desc="Example parameter (units)")
    target_on_time = traits.Float(1.0, desc="Time to display target")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target = VirtualCircularTarget(target_radius=1, target_color=[1,1,1,1])
        for model in self.target.graphics_models:
            self.add_model(model)

    def init(self):
        self.add_dtype('example_cycle_data_to_save', 'f8', (1,))
        self.trial_dtype = np.dtype([('trial', 'u4'), ('index', 'u4'), ('target', 'f8', (3,))])
        super().init()

    def _cycle(self):
        '''
        Calls any update functions necessary and redraws screen
        '''
        super()._cycle()

    #### TEST FUNCTIONS ####
    def _test_start_trial(self, ts):
        return not self.pause

    def _test_start_pause(self, ts):
        return self.pause
    
    def _test_target_done(self, ts):
        return ts > self.target_on_time
    
    def _test_end_pause(self, ts):
        return not self.pause

    def _parse_next_trial(self):
        '''Check that the generator has the required data'''
        self.target_idx, self.target_location = self.next_trial
        
        # Update the data sinks with trial information
        self.trial_record['trial'] = self.calc_trial_num()
        self.trial_record['index'] = self.target_idx
        self.trial_record['target'] =  self.target_location
        self.sinks.send("trials", self.trial_record)

    ### STATE FUNCTIONS ###
    def _start_wait(self):
        self.sync_event('TARGET_OFF')
        self.target.hide()
        super()._start_wait()

    def _start_target(self):
        self.sync_event('TARGET_ON', 0xd)
        
        self.task_data['example_cycle_data_to_save'] = 2 # maybe you want to save some specific data here
        
        print("driving target to", self.target_location)
        self.target.move_to_position(self.target_location)
        self.target.show()

    def _start_pause(self):
        self.sync_event('PAUSE_START')
        self.target.hide()

    def _end_pause(self):
        self.sync_event('PAUSE_END')

    @classmethod
    def get_desc(cls, params, log_summary):
        duration = round(log_summary['runtime'] / 60, 1)
        return "{}/{} succesful trials in {} min".format(
            log_summary['n_success_trials'], log_summary['n_trials'], duration)

    @staticmethod
    def example_generator():
        '''
        Generates a sequence of targets

        Returns
        -------
        [10 x 1] array of tuples containing trial index and [1 x 3] target coordinates

        '''
        
        for idx in range(10):
            pos = np.array([idx,0,1]).T
            yield idx, pos




'''
DIO using a serial port + microcontroller instead of the NIDAQ card
'''
import serial
from collections import defaultdict
import struct
from numpy import binary_repr
from .dio.parse import MSG_TYPE_ROWBYTE, MSG_TYPE_REGISTER

def construct_word(aux, msg_type, data, n_bits_data=8, n_bits_msg_type=3):
    word = (aux << (n_bits_data + n_bits_msg_type)) | (msg_type << n_bits_data) | data
    return word

def parse_word(word, n_bits_data=8, n_bits_msg_type=3):
    data = word & ((1 << n_bits_data) - 1)
    msg_type = (word >> n_bits_data) & ((1 << n_bits_msg_type) - 1) 
    aux = word >> n_bits_msg_type + n_bits_data
    return aux, msg_type, data

baudrate = 115200

class SendRowByte(object):
    '''
    Send only an 8-bit data word corresponding to the 8 lower 
    bits of the current row number of the HDF table
    '''
    '''
    Interface for sending all the task-generated data through the NIDAQ interface card
    '''
    def __init__(self, device=None):
        '''
        Constructor for SendRowByte

        Parameters
        ----------
        device : string, optional
            Linux name of the serial port for the Arduino board, defined by setserial

        Returns
        -------
        SendAll instance
        '''
        self.systems = dict()
        self.port = serial.Serial('/dev/arduino_neurosync', baudrate=baudrate)
        self.n_systems = 0
        self.rowcount = defaultdict(int)

    def close(self):
        '''
        Release access to the Arduino serial port
        '''
        # stop recording
        self.port.write('p')
        self.port.close()

    def register(self, system, dtype):
        '''
        Send information about the registration system (name and datatype) in string form, one byte at a time.

        Parameters
        ----------
        system : string
            Name of the system being registered
        dtype : np.dtype instance
            Datatype of incoming data, for later decoding of the binary data during analysis

        Returns
        -------
        None
        '''
        # Save the index of the system being registered (arbitrary number corresponding to the order in which systems were registered)
        self.n_systems += 1
        self.systems[system] = self.n_systems


        # if self.n_systems > 1:
        #     raise Exception("This currently only works for one system!")
        
        #print "System Register: %s" % system, self.systems[system]
        #print "Arduino register %s" % system, self.systems[system]

        #if self.n_systems > 1:
        #    raise Exception("This currently only works for one system!")

        print("Arduino register %s" % system, self.systems[system])

        for sys_name_chr in system:
            reg_word = construct_word(self.systems[system], MSG_TYPE_REGISTER, ord(sys_name_chr))
            self._send_data_word_to_serial_port(reg_word)

        null_term_word = construct_word(self.systems[system], MSG_TYPE_REGISTER, 0) # data payload is 0 for null terminator
        self._send_data_word_to_serial_port(null_term_word)

    def sendMsg(self, msg):
        '''
        Do nothing. Messages are stored with row numbers in the HDF table, so no need to also send the message over to the recording system.

        Parameters
        ----------
        msg : string
            Message to send

        Returns
        -------
        None
        '''
        # there's no point in sending a message, since every message is 
        # stored in the HDF table anyway with a row number, 
        # and every row number is automatically synced.
        pass

    def send(self, system, data):
        '''
        Send the row number for a data word to the neural system

        Parameters
        ----------
        system : string 
            Name of system 
        data : object
            This is unused. Only used in the parent's version where the actual data, and not just the HDF row number, is sent.

        Returns
        -------
        None
        '''
        
        if not (system in self.systems):
            # if the system is not registered, do nothing
            return

        current_sys_rowcount = self.rowcount[system]
        self.rowcount[system] += 1

        # construct the data packet
        word = construct_word(self.systems[system], MSG_TYPE_ROWBYTE, current_sys_rowcount % 256)
        self._send_data_word_to_serial_port(word)
        
        # if verbose:
        #     print binary_repr(word, 16)
        # word_str = 'd' + struct.pack('<H', word)
        # self.port.write(word_str)

    def _send_data_word_to_serial_port(self, word, verbose=False):
        #self.port.write(word)
        if verbose:
            print(binary_repr(word, 16))
        word_str = 'd' + struct.pack('<H', word)
        self.port.write(word_str)



'''
Code for interacting with the Crist reward system(s). Consult the Crist manual for the command protocol
'''
import time

import struct
import binascii
import threading
import io
import traceback


import serial
import time


try:
    import traits.api as traits
except:
    import enthought.traits.api as traits

def _xsum(msg):
    '''
    Compute the checksums for the messages which must be sent as part of the packet

    Parameters
    ----------
    msg : string
        Message to be sent over the serial port

    Returns
    -------
    char
        The 8-bit checksum of the entire message
    '''
    chrval = [int(''.join(x), 16) for x in zip(*[iter(binascii.b2a_hex(msg))]*2)]
    return chr(sum(chrval) % 256)

class Basic(object):
    '''
    Bare-bones interface for the Crist reward system. Can give timed reward and drain on/off.
    This class is sufficient for all the tasks implemented as of Aug. 2014.
    '''
    response_message_length = 7
    def __init__(self):
        '''
        Constructor for basic reward system interface

        Parameters
        ----------
        None

        Returns
        -------
        Basic instance
        '''
        self.port = serial.Serial('/dev/crist_reward', baudrate=38400)
        from config import config
        self.version = int(config.reward_sys['version'])
        if self.version==1: self.set_beeper_volume(128)
        time.sleep(.5)
        self.reset()

    def _write(self, msg):
        '''
        Send an arbitrary message over the serial port

        Parameters
        ----------
        msg : string
            Message to be sent over the serial port        

        Returns
        -------
        msg_out : string
            Response from crist system after sending command
        '''
        fmsg = msg+_xsum(msg)
        self.port.flushOutput()
        self.port.flushInput()
        self.port.write(fmsg)
        msg_out = self.port.read(self.port.inWaiting())
        return msg_out

    def reward(self, length):
        '''
        Open the solenoid for some length of time

        Parameters
        ----------
        length : float
            Duration of time the solenoid should be open, in seconds. NOTE: in some versions of the system, there appears to be max of ~5s

        Returns
        -------
        None
        '''
        length /= .1
        length = int(length)
        if self.version==0:
            self._write(struct.pack('<ccxHxx', '@', 'G', length))
        elif self.version==1:
            self._write(struct.pack('<cccHxxx', '@', 'G', '1', length))
        else:
            raise Exception("Unrecognized reward system version!")
        self.port.read(self.port.inWaiting())
    
    def setup_touch_sensor(self):
        '''
        Send the serial command to initialize the Crist touch sensor
        '''
        if self.version==1: #arc system
            cmd = ['@', 'C', '1' ,'O','%c' % 0b10000000, '%c' % 1, '%c' % 0, 'E']
            stuff = ''.join(cmd)
            self._write(stuff)

    def sensor_reward(self, length):
        '''
        Set the duration of the reward if the subject touches the touch sensor

        Parameters
        ----------
        length : float
            Duration of time the solenoid should be open, in seconds. NOTE: in some versions of the system, there appears to be max of ~5s

        Returns
        -------
        None
        '''
        if self.version==1:
            cmd = ['@', 'S',  '%c' % 0x02, '%c' % 0x02, '%c' %10, '%c' %0, '%c' %0]
            stuff = ''.join(cmd)
            self._write(stuff)

    def set_beeper_volume(self, volume):
        '''
        Send a command to set the sound level of the audio beep paired with the solenoid opening

        Parameters
        ----------
        volume : int in range [0, 255]
            255 is max possible volume

        Returns
        -------
        string 
            Response message from system
        '''
        if not (volume >= 0 and volume <= 255):
            raise ValueError("Invalid beeper volume: %g" % volume)
        return self._write('@CS' + '%c' % volume + 'E' + struct.pack('xxx')) 

    def reset(self):
        '''
        Send the system reset command
        '''
        if self.version==0:
            self._write("@CPSNNN")
        elif self.version==1:
            cmd = ['@', 'C', '1', 'P', '%c' % 0b10000000, '%c' % 0, '%c' % 0, 'D']
            stuff = ''.join(cmd)
            self._write(stuff)
        else:
            raise Exception("Unrecognized reward system version!")
        self.last_response = self.port.read(self.port.inWaiting())

    def drain(self, drain_time=1200):
        '''
        Turns on the reward system drain for specified amount of time (in seconds)

        Parameters
        ----------
        drain_time : float 
            Time to drain the system, in seconds.

        Returns
        -------
        None
        '''
        assert drain_time > 0
        assert drain_time < 9999
        if self.version == 0: #have to wait and manually tell it to turn off
            self._write("@CNSENN")
            time.sleep(drain_time)
            self._write("@CNSDNN")
        elif self.version == 1:
            self._write('@M1' + struct.pack('H', drain_time) + 'D' + struct.pack('xx'))
        else:
            raise Exception("Unrecognized reward system version!")

    def drain_off(self):
        '''
        Turns off drain if currently on
        '''
        if self.version==0:
            self._write("@CNSDNN")
        elif self.version==1:
            self._write('@M1' + struct.pack('H', 0) + 'A' + struct.pack('xx'))
        else:
            raise Exception("Unrecognized reward system version!")


##########################################
##### Code below this line is unused #####
##########################################
class _parse_num(object):
    types = {1:'<B', 2:'<H', 4:'<I'}
    def __init__(self, length=2, mult=1, unit=""):
        self.t = self.types[length]
        self.m = mult
        self.u = unit
    def __getitem__(self, msg):
        if len(msg) == 3:
            msg += "\x00"
        i, = struct.unpack(self.t, msg)
        return i*self.m

class System(traits.HasTraits, threading.Thread):
    '''
    More complete reward system interface. Only tested for "version 0" of the system
    '''
    _running = True
    port = traits.Instance(serial.Serial)

    reward_mode = traits.Enum("Time", "Volume", "Count")
    sensor_status = traits.Bool
    drain_status = traits.Enum("Disabled", "PC enabled", "Switch enabled", "External enabled", "Time out")
    ctrl_status = traits.Enum("Run", "Halted high", "Halted low", "Offline", "Halted problem", "Override timeout")
    switch_state = traits.Enum("0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "A", "B", "C", "D", "E", "F")
    reward_freq = traits.Float
    touches_per_reward = traits.Int
    programmed_time = traits.Float
    programmed_volume = traits.Float
    total_rewards = traits.Int
    total_touches = traits.Int
    total_reward_time = traits.Float
    total_reward_volume = traits.Float
    sensor_id = traits.Str
    dispenser_id = traits.Str
    firmware_version = traits.Float

    def __init__(self, **kwargs):
        super(System, self).__init__(**kwargs)
        threading.Thread.__init__(self)
        self.plock = threading.Lock()
        self.daemon = True
        self._messages = dict([
            ("&D",(37, "Data", self._parse_status)), 
            ("#E", (4, "Error")), 
            ("#A", (4, "Acknowledge")), 
            ("*Z", (8, "Volume Calibration"))
        ])
        reward_mode = dict(T="Time", V="Volume", C="Count")
        drain_status = dict(
            P="PC enabled", 
            S="Switch enabled", 
            X="External enabled", 
            D="Disabled", 
            T="Time out")
        ctrl_status = dict(
            R="Run", 
            H="halted, input voltage high", 
            L="halted, input voltage low", 
            D="Controller device off-line", 
            M="Halted, message problem",
            O="override timeout")
        enable_disable = dict(
            D=False,
            E=True)
        self._order = [
            (1, "reward_mode", reward_mode),
            (1, "sensor_status", enable_disable),
            (1, "drain_status", drain_status),
            (1, "ctrl_status", ctrl_status),
            (1, "switch_state", None),
            (2, "reward_freq", _parse_num(unit="s")),
            (1, "touches_per_reward", _parse_num(1)),
            (2, "programmed_time", _parse_num(2, 0.1, "ms")),
            (2, "programmed_volume", _parse_num(2, .001, "ml")),
            (2, "total_rewards", _parse_num()),
            (2, "total_touches", _parse_num()),
            (4, "total_reward_time", _parse_num(4, 0.1, "ms")),
            (3, "total_reward_volume", _parse_num(4, .001, "ml")),
            (5, "sensor_ID", None),
            (5, "dispenser_ID", None),
            (1, "firmware_version", _parse_num(1, .1))
            ]
        #self.reset_stats()

    def _parse_status(self, msg):
        msg = io.StringIO(msg)
        output = {}
        for length, name, op in self._order:
            part = msg.read(length)
            if op is None:
                output[name] = part
            else:
                output[name] = op[part]
        
        self.set(**output)

    def __del__(self):
        self._running = False
    
    def run(self):
        while self._running:
            header = self.port.read(2)
            print("recieved %r"%header)
            try:
                self.plock.acquire()
                msg = self.port.read(self._messages[header][0] - 2)
                self.plock.release()
                assert _xsum(header+msg[:-1]) == msg[-1], "Wrong checksum! %s"%msg
                if len(self._messages[header]) > 2:
                    self._messages[header][-1](msg)
                else:
                    print(self._messages[header], repr(msg))
            except:
                traceback.print_exc()
                time.sleep(10)
                print(repr(msg),repr(self.port.read(self.port.inWaiting())))
    
    def reward(self, time=500, volume=None):
        '''Returns the string used to output a time or volume reward.
        
        Parameters
        ----------
        time : int
            Time in milliseconds to turn on the reward
        volume: int
            volume in microliters
        '''
        assert (volume is None and time is not None) or \
            (volume is not None and time is None)
        time /= .1
        self._write(struct.pack('<ccxHxx', '@', 'G', time))
    
    def _write(self, msg):
        fmsg = msg+_xsum(msg)
        self.plock.acquire()
        self.port.write(fmsg)
        self.plock.release()
    
    def update(self):
        self._write("@CNSNNN")
    
    def reset(self):
        self._write("@CPSNNN")
    
    def reset_stats(self):
        self._write("@CRSNNN")
    
    def drain(self, status=None):
        mode = ("D", "E")[self.drain_status == "Disabled" if status is None else status]
        self._write("@CNS%sNN"%mode)

def open():
    try:
        #port = serial.Serial(glob.glob("/dev/ttyUSB*")[0], baudrate=38400)
        #reward = System(port=port)
        #reward.start()
        reward = Basic()
        return reward
    except:
        print("Reward system not found")
        import traceback
        import os
        import builtins
        traceback.print_exc(file=builtins.open(os.path.expanduser('~/code/bmi3d/log/reward.log'), 'w'))



"""
Primitive operations for 3x3 orthonormal and 4x4 homogeneous matrices.

@author: Peter Corke
@copyright: Peter Corke
"""

from numpy import *
from .utility import *
from numpy.linalg import norm
from .Quaternion import *


def rotx(theta):
    """
    Rotation about X-axis
    
    @type theta: number
    @param theta: the rotation angle
    @rtype: 3x3 orthonormal matrix
    @return: rotation about X-axis

    @see: L{roty}, L{rotz}, L{rotvec}
    """
    
    ct = cos(theta)
    st = sin(theta)
    return mat([[1,  0,    0],
            [0,  ct, -st],
            [0,  st,  ct]])

def roty(theta):
    """
    Rotation about Y-axis
    
    @type theta: number
    @param theta: the rotation angle
    @rtype: 3x3 orthonormal matrix
    @return: rotation about Y-axis

    @see: L{rotx}, L{rotz}, L{rotvec}
    """
    
    ct = cos(theta)
    st = sin(theta)

    return mat([[ct,   0,   st],
            [0,    1,    0],
            [-st,  0,   ct]])

def rotz(theta):
    """
    Rotation about Z-axis
    
    @type theta: number
    @param theta: the rotation angle
    @rtype: 3x3 orthonormal matrix
    @return: rotation about Z-axis

    @see: L{rotx}, L{roty}, L{rotvec}
    """
    
    ct = cos(theta)
    st = sin(theta)

    return mat([[ct,      -st,  0],
            [st,       ct,  0],
            [ 0,    0,  1]])

def trotx(theta):
    """
    Rotation about X-axis
    
    @type theta: number
    @param theta: the rotation angle
    @rtype: 4x4 homogeneous matrix
    @return: rotation about X-axis

    @see: L{troty}, L{trotz}, L{rotx}
    """
    return r2t(rotx(theta))

def troty(theta):
    """
    Rotation about Y-axis
    
    @type theta: number
    @param theta: the rotation angle
    @rtype: 4x4 homogeneous matrix
    @return: rotation about Y-axis

    @see: L{troty}, L{trotz}, L{roty}
    """
    return r2t(roty(theta))

def trotz(theta):
    """
    Rotation about Z-axis
    
    @type theta: number
    @param theta: the rotation angle
    @rtype: 4x4 homogeneous matrix
    @return: rotation about Z-axis

    @see: L{trotx}, L{troty}, L{rotz}
    """
    return r2t(rotz(theta))


##################### Euler angles

def tr2eul(m):
    """
    Extract Euler angles.
    Returns a vector of Euler angles corresponding to the rotational part of 
    the homogeneous transform.  The 3 angles correspond to rotations about
    the Z, Y and Z axes respectively.
    
    @type m: 3x3 or 4x4 matrix
    @param m: the rotation matrix
    @rtype: 1x3 matrix
    @return: Euler angles [S{theta} S{phi} S{psi}]
    
    @see:  L{eul2tr}, L{tr2rpy}
    """
    
    try:
        m = mat(m)
        if ishomog(m):
            euler = mat(zeros((1,3)))
            if norm(m[0,2])<finfo(float).eps and norm(m[1,2])<finfo(float).eps:
                # singularity
                euler[0,0] = 0
                sp = 0
                cp = 1
                euler[0,1] = arctan2(cp*m[0,2] + sp*m[1,2], m[2,2])
                euler[0,2] = arctan2(-sp*m[0,0] + cp*m[1,0], -sp*m[0,1] + cp*m[1,1])
                return euler
            else:
                euler[0,0] = arctan2(m[1,2],m[0,2])
                sp = sin(euler[0,0])
                cp = cos(euler[0,0])
                euler[0,1] = arctan2(cp*m[0,2] + sp*m[1,2], m[2,2])
                euler[0,2] = arctan2(-sp*m[0,0] + cp*m[1,0], -sp*m[0,1] + cp*m[1,1])
                return euler
            
    except ValueError:
        euler = []
        for i in range(0,len(m)):
            euler.append(tr2eul(m[i]))
        return euler
        

def eul2r(phi, theta=None, psi=None):
    """
    Rotation from Euler angles.
    
    Two call forms:
        - R = eul2r(S{theta}, S{phi}, S{psi})
        - R = eul2r([S{theta}, S{phi}, S{psi}])
    These correspond to rotations about the Z, Y, Z axes respectively.

    @type phi: number or list/array/matrix of angles
    @param phi: the first Euler angle, or a list/array/matrix of angles
    @type theta: number
    @param theta: the second Euler angle
    @type psi: number
    @param psi: the third Euler angle
    @rtype: 3x3 orthonormal matrix
    @return: R([S{theta} S{phi} S{psi}])

    @see:  L{tr2eul}, L{eul2tr}, L{tr2rpy}

    """

    n = 1
    if theta == None and psi==None:
        # list/array/matrix argument
        phi = mat(phi)
        if numcols(phi) != 3:
            error('bad arguments')
        else:
            n = numrows(phi)
            psi = phi[:,2]
            theta = phi[:,1]
            phi = phi[:,0]
    elif (theta!=None and psi==None) or (theta==None and psi!=None):
        error('bad arguments')
    elif not isinstance(phi,(int,int32,float,float64)):
        # all args are vectors
        phi = mat(phi)
        n = numrows(phi)
        theta = mat(theta)
        psi = mat(psi)

    if n>1:
        R = []
        for i in range(0,n):
                r = rotz(phi[i,0]) * roty(theta[i,0]) * rotz(psi[i,0])
                R.append(r)
        return R
    try:
        r = rotz(phi[0,0]) * roty(theta[0,0]) * rotz(psi[0,0])
        return r
    except:
        r = rotz(phi) * roty(theta) * rotz(psi)
        return r

def eul2tr(phi,theta=None,psi=None):
    """
    Rotation from Euler angles.
    
    Two call forms:
        - R = eul2tr(S{theta}, S{phi}, S{psi})
        - R = eul2tr([S{theta}, S{phi}, S{psi}])
    These correspond to rotations about the Z, Y, Z axes respectively.

    @type phi: number or list/array/matrix of angles
    @param phi: the first Euler angle, or a list/array/matrix of angles
    @type theta: number
    @param theta: the second Euler angle
    @type psi: number
    @param psi: the third Euler angle
    @rtype: 4x4 homogenous matrix
    @return: R([S{theta} S{phi} S{psi}])

    @see:  L{tr2eul}, L{eul2r}, L{tr2rpy}

    """
    return r2t( eul2r(phi, theta, psi) )


################################## RPY angles


def tr2rpy(m,zyx=False):
    """
    Extract RPY angles.
    Returns a vector of RPY angles corresponding to the rotational part of 
    the homogeneous transform.  The 3 angles correspond to rotations about
    the Z, Y and X axes respectively.
    
    @type m: 3x3 or 4x4 matrix
    @param m: the rotation matrix
    @rtype: 1x3 matrix
    @return: RPY angles [S{theta} S{phi} S{psi}]
    
    @see:  L{rpy2tr}, L{tr2eul}
    """
    try:
        if ~zyx:
            
            m = mat(m)
            if ishomog(m):
                rpy = mat(zeros((1,3)))
                if norm(m[2,2])<finfo(float).eps and norm(m[1,2])<finfo(float).eps:
                    # singularity
                    rpy[0,0] = 0
                    rpy[0,1] = arctan2(m[0,2], m[2,2]) # pitch
                    rpy[0,2] = arctan2(m[1,0], m[1,1]) # yaw is sum of roll+yaw
                    return rpy
                else:
                    rpy[0,0] = arctan2(-m[1,3],m[2,2])
                    sp = sin(rpy[0,0])
                    cp = cos(rpy[0,0])
                    rpy[0,1] = arctan2(m[0,2], cp*m[2,2] - sp*m[1,2])# pitch
                    rpy[0,2] = arctan2(-m[0,1], m[0,0])# yaw
                    return rpy
        else:
            m = mat(m)
            if ishomog(m):
                rpy = mat(zeros((1,3)))
                if norm(m[0,0])<finfo(float).eps and norm(m[1,0])<finfo(float).eps:
                    # singularity
                    rpy[0,0] = 0
                    rpy[0,1] = arctan2(-m[2,0], m[0,0])
                    rpy[0,2] = arctan2(-m[1,2], m[1,1])
                    return rpy
                else:
                    rpy[0,0] = arctan2(m[1,0],m[0,0])
                    sp = sin(rpy[0,0])
                    cp = cos(rpy[0,0])
                    rpy[0,1] = arctan2(-m[2,0], cp*m[0,0] + sp*m[1,0])
                    rpy[0,2] = arctan2(sp*m[0,2] - cp*m[1,2], cp*m[1,1] - sp*m[0,1])
                    return rpy    

    except ValueError:
        rpy = []
        for i in range(0,len(m)):
            rpy.append(tr2rpy(m[i]))
        return rpy
        
def rpy2r(roll, pitch=None,yaw=None,zyx=False,deg=False):
    """
    Rotation from RPY angles.
    
    Two call forms:
        - R = rpy2r(S{theta}, S{phi}, S{psi})
        - R = rpy2r([S{theta}, S{phi}, S{psi}])
    These correspond to rotations about the Z, Y, X axes respectively.

    @type roll: number or list/array/matrix of angles
    @param roll: roll angle, or a list/array/matrix of angles
    @type pitch: number
    @param pitch: pitch angle
    @type yaw: number
    @param yaw: yaw angle
    @rtype: 4x4 homogenous matrix
    @return: R([S{theta} S{phi} S{psi}])

    @see:  L{tr2rpy}, L{rpy2r}, L{tr2eul}

    """
    n=1
    if pitch==None and yaw==None:
        roll= mat(roll)
        if numcols(roll) != 3:
            error('bad arguments')
        n = numrows(roll)
        pitch = roll[:,1]
        yaw = roll[:,2]
        roll = roll[:,0]

    if deg:
        #convert to degrees 
        d2r = pi/180.0
        roll = roll * d2r;
        pitch = pitch * d2r;
        yaw = yaw * d2r;
        
    if ~zyx:
        # XYZ order
        if n>1:
            R = []
            for i in range(0,n):
                r = rotz(roll[i,0]) * roty(pitch[i,0]) * rotx(yaw[i,0])
                R.append(r)
            return R
        try:
            r = rotz(roll[0,0]) * roty(pitch[0,0]) * rotx(yaw[0,0])
            return r
        except:
            r = rotx(roll) * roty(pitch) * rotz(yaw)
            return r
    else:
        # XYZ order
        if n>1:
            R = []
            for i in range(0,n):
                r = rotz(roll[i,0]) * roty(pitch[i,0]) * rotx(yaw[i,0])
                R.append(r)
            return R
        try:
            r = rotz(roll[0,0]) * roty(pitch[0,0]) * rotx(yaw[0,0])
            return r
        except:
            r = rotz(roll) * roty(pitch) * rotx(yaw)
            return r


def rpy2tr(roll, pitch=None, yaw=None, zyx=False, deg=False):
    """
    Rotation from RPY angles.
    
    Two call forms:
        - R = rpy2tr(r, p, y)
        - R = rpy2tr([r, p, y])
    These correspond to rotations about the Z, Y, X axes respectively.

    @type roll: number or list/array/matrix of angles
    @param roll: roll angle, or a list/array/matrix of angles
    @type pitch: number
    @param pitch: pitch angle
    @type yaw: number
    @param yaw: yaw angle
    @rtype: 4x4 homogenous matrix
    @return: R([S{theta} S{phi} S{psi}])

    @see:  L{tr2rpy}, L{rpy2r}, L{tr2eul}

    """
    return r2t( rpy2r(roll, pitch, yaw, zxy, deg) )

###################################### OA vector form


def oa2r(o,a):
    """Rotation from 2 vectors.
    The matrix is formed from 3 vectors such that::
        R = [N O A] and N = O x A.  

    In robotics A is the approach vector, along the direction of the robot's 
    gripper, and O is the orientation vector in the direction between the 
    fingertips.
    
    The submatrix is guaranteed to be orthonormal so long as O and A are 
    not parallel.
    
    @type o: 3-vector
    @param o: The orientation vector.
    @type a: 3-vector
    @param a: The approach vector
    @rtype: 3x3 orthonormal rotation matrix
    @return: Rotatation matrix
    
    @see: L{rpy2r}, L{eul2r}
    """
    n = crossp(o, a)
    n = unit(n)
    o = crossp(a, n);
    o = unit(o).reshape(3,1)
    a = unit(a).reshape(3,1)
    return bmat('n o a')


def oa2tr(o,a):
    """otation from 2 vectors.
    The rotation submatrix is formed from 3 vectors such that::

        R = [N O A] and N = O x A.  

    In robotics A is the approach vector, along the direction of the robot's 
    gripper, and O is the orientation vector in the direction between the 
    fingertips.
    
    The submatrix is guaranteed to be orthonormal so long as O and A are 
    not parallel.
    
    @type o: 3-vector
    @param o: The orientation vector.
    @type a: 3-vector
    @param a: The approach vector
    @rtype: 4x4 homogeneous transformation matrix
    @return: Transformation matrix
    
    @see: L{rpy2tr}, L{eul2tr}
    """
    return r2t(oa2r(o,a))
    
    
###################################### angle/vector form


def rotvec2r(theta, v):
    """
    Rotation about arbitrary axis.  Compute a rotation matrix representing
    a rotation of C{theta} about the vector C{v}.
    
    @type v: 3-vector
    @param v: rotation vector
    @type theta: number
    @param theta: the rotation angle
    @rtype: 3x3 orthonormal matrix
    @return: rotation

    @see: L{rotx}, L{roty}, L{rotz}
    """
    v = arg2array(v);
    ct = cos(theta)
    st = sin(theta)
    vt = 1-ct
    r = mat([[ct,         -v[2]*st,    v[1]*st],\
             [v[2]*st,          ct,   -v[0]*st],\
             [-v[1]*st,  v[0]*st,           ct]])
    return v*v.T*vt+r

def rotvec2tr(theta, v):
    """
    Rotation about arbitrary axis.  Compute a rotation matrix representing
    a rotation of C{theta} about the vector C{v}.
    
    @type v: 3-vector
    @param v: rotation vector
    @type theta: number
    @param theta: the rotation angle
    @rtype: 4x4 homogeneous matrix
    @return: rotation

    @see: L{trotx}, L{troty}, L{trotz}
    """
    return r2t(rotvec2r(theta, v))


###################################### translational transform


def transl(x, y=None, z=None):
    """
    Create or decompose translational homogeneous transformations.
    
    Create a homogeneous transformation
    ===================================
    
        - T = transl(v)
        - T = transl(vx, vy, vz)
        
        The transformation is created with a unit rotation submatrix.
        The translational elements are set from elements of v which is
        a list, array or matrix, or from separate passed elements.
    
    Decompose a homogeneous transformation
    ======================================
    

        - v = transl(T)   
    
        Return the translation vector
    """
           
    if y==None and z==None:
            x=mat(x)
            try:
                    if ishomog(x):
                            return x[0:3,3].reshape(3,1)
                    else:
                            return concatenate((concatenate((eye(3),x.reshape(3,1)),1),mat([0,0,0,1])))
            except AttributeError:
                    n=len(x)
                    r = [[],[],[]]
                    for i in range(n):
                            r = concatenate((r,x[i][0:3,3]),1)
                    return r
    elif y!=None and z!=None:
            return concatenate((concatenate((eye(3),mat([x,y,z]).T),1),mat([0,0,0,1])))


###################################### Skew symmetric transform


def skew(*args):
    """
    Convert to/from skew-symmetric form.  A skew symmetric matrix is a matrix
    such that M = -M'
    
    Two call forms
    
        -ss = skew(v)
        -v = skew(ss)
        
    The first form builds a 3x3 skew-symmetric from a 3-element vector v.
    The second form takes a 3x3 skew-symmetric matrix and returns the 3 unique
    elements that it contains.
    
    """
    
    def ss(b):
        return  matrix([
            [0, -b[2],  b[1]],
            [b[2],  0,  -b[0]],
            [-b[1], b[0],   0]]);

    if len(args) == 1:
        # convert matrix to skew vector
        b = args[0];
        
        if isrot(b):
            return 0.5*matrix( [b[2,1]-b[1,2], b[0,2]-b[2,0], b[1,0]-b[0,1]] );
        elif ishomog(b):
            return vstack( (b[0:3,3], 0.5*matrix( [b[2,1]-b[1,2], b[0,2]-b[2,0], b[1,0]-b[0,1]] ).T) );

    
    # build skew-symmetric matrix
          
        b = arg2array(b);
        if len(b) == 3:
            return ss(b);
        elif len(b) == 6:
            r = hstack( (ss(b[3:6]), mat(b[0:3]).T) );
            r = vstack( (r, mat([0, 0, 0, 1])) );
            return r;
            
    elif len(args) == 3:
            return ss(args);    
    elif len(args) == 6:
            r = hstack( (ss(args[3:6]), mat(args[0:3]).T) );
            r = vstack( (r, mat([0, 0, 0, 1])) );
            return r;    
    else:
        raise ValueError;



def tr2diff(t1, t2):
    """
    Convert a transform difference to differential representation.
    Returns the 6-element differential motion required to move
    from T1 to T2 in base coordinates.
    
    @type t1: 4x4 homogeneous transform
    @param t1: Initial value
    @type t2: 4x4 homogeneous transform
    @param t2: Final value
    @rtype: 6-vector
    @return: Differential motion [dx dy dz drx dry drz]
    @see: L{skew}
    """
    
    t1 = mat(t1)
    t2 = mat(t2)
    
    d = concatenate(
        (t2[0:3,3]-t1[0:3,3],
        0.5*(   crossp(t1[0:3,0], t2[0:3,0]) +
                crossp(t1[0:3,1], t2[0:3,1]) +
                crossp(t1[0:3,2], t2[0:3,2]) )
            ))
    return d

################################## Utility


def trinterp(T0, T1, r):
    """
    Interpolate homogeneous transformations.
    Compute a homogeneous transform interpolation between C{T0} and C{T1} as
    C{r} varies from 0 to 1 such that::
    
        trinterp(T0, T1, 0) = T0
        trinterp(T0, T1, 1) = T1
        
    Rotation is interpolated using quaternion spherical linear interpolation.

    @type T0: 4x4 homogeneous transform
    @param T0: Initial value
    @type T1: 4x4 homogeneous transform
    @param T1: Final value
    @type r: number
    @param r: Interpolation index, in the range 0 to 1 inclusive
    @rtype: 4x4 homogeneous transform
    @return: Interpolated value
    @see: L{quaternion}, L{ctraj}
    """
    
    q0 = Quaternion(T0)
    q1 = Quaternion(T1)
    p0 = transl(T0)
    p1 = transl(T1)


    qr = q0.interp(q1, r)
    pr = p0*(1-r) + r*p1

    return vstack( (concatenate((qr.R(),pr),1), mat([0,0,0,1])) )


def trnorm(t):
    """
    Normalize a homogeneous transformation.
    Finite word length arithmetic can cause transforms to become `unnormalized',
    that is the rotation submatrix is no longer orthonormal (det(R) != 1).
    
    The rotation submatrix is re-orthogonalized such that the approach vector
    (third column) is unchanged in direction::
    
        N = O x A
        O = A x N

    @type t: 4x4 homogeneous transformation
    @param t: the transform matrix to convert
    @rtype: 3x3 orthonormal rotation matrix
    @return: rotation submatrix
    @see: L{oa2tr}
    @bug: Should work for 3x3 matrix as well.
    """

    t = mat(t)      # N O A
    n = crossp(t[0:3,1],t[0:3,2]) # N = O X A
    o = crossp(t[0:3,2],t[0:3,0]) # O = A x N
    return concatenate(( concatenate((unit(n),unit(t[0:3,1]),unit(t[0:3,2]),t[0:3,3]),1),
         mat([0,0,0,1])))


def t2r(T):
    """
    Return rotational submatrix of a homogeneous transformation.
    @type T: 4x4 homogeneous transformation
    @param T: the transform matrix to convert
    @rtype: 3x3 orthonormal rotation matrix
    @return: rotation submatrix
    """    
    
    if ishomog(T)==False:
        error( 'input must be a homogeneous transform')
    return T[0:3,0:3]


def r2t(R):
    """
    Convert a 3x3 orthonormal rotation matrix to a 4x4 homogeneous transformation::
    
        T = | R 0 |
            | 0 1 |
            
    @type R: 3x3 orthonormal rotation matrix
    @param R: the rotation matrix to convert
    @rtype: 4x4 homogeneous matrix
    @return: homogeneous equivalent
    """
    
    return concatenate( (concatenate( (R, zeros((3,1))),1), mat([0,0,0,1])) )



'''
Classes for BMI decoding using the Point-process filter. 
'''

import numpy as np

from . import bmi
from .bmi import GaussianState
import cmath

class PointProcessFilter(bmi.GaussianStateHMM):
    """
    Low-level Point-process filter, agnostic to application

    Model: 
       x_{t+1} = Ax_t + Bu_t + w_t; w_t ~ N(0, W)
       log(y_t) = Cx_t

    See Shanechi et al., "Feedback-Controlled Parallel Point Process Filter for 
    Estimation of Goal-Directed Movements From Neural Signals", IEEE TNSRE, 2013
    for mathematical details.
    """
    model_attrs = ['A', 'W', 'C']

    def __init__(self, A=None, W=None, C=None, dt=None, is_stochastic=None, B=0, F=0):
        '''
        Constructor for PointProcessFilter

        Parameters
        ----------
        A : np.mat
            Model of state transition matrix
        W : np.mat
            Model of process noise covariance
        C : np.mat
            Model of conditional distribution between observations and hidden state
            log(obs) = C * hidden_state
        dt : float
            Discrete-time sampling rate of the filter. Used to map spike counts to spike rates
        B : np.ndarray, optional
            Control input matrix
        F : np.ndarray, optional
            State-space feedback gain matrix to drive state back to equilibrium state.
        is_stochastic : np.array, optional
            Array of booleans specifying for each state whether it is stochastic. 
            If 'None' specified, all states are assumed to be stochastic

        Returns
        -------
        KalmanFilter instance
        '''
        if A is None and W is None and C is None and dt is None:
            ## This condition should only be true in the unpickling phase
            pass            
        else:
            self.A = np.mat(A)
            self.W = np.mat(W)
            self.C = np.mat(C)
            self.dt = dt
            self.spike_rate_dt = dt
            
            self.B = B
            self.F = F

            if is_stochastic == None:
                n_states = A.shape[0]
                self.is_stochastic = np.ones(n_states, dtype=bool)
            else:
                self.is_stochastic = np.array(is_stochastic)
            
            self.state_noise = GaussianState(0.0, W)
            self._pickle_init()

    def _pickle_init(self):
        """
        Code common to unpickling and initialization
        """
        nS = self.A.shape[0]
        offset_row = np.zeros(nS)
        offset_row[-1] = 1
        self.include_offset = np.array_equal(np.array(self.A)[-1, :], offset_row)

        self.spike_rate_dt = self.dt

        if not hasattr(self, 'B'): self.B = 0
        if not hasattr(self, 'F'): self.F = 0

    def init_noise_models(self):
        '''
        see bmi.GaussianStateHMM.init_noise_models for documentation
        '''
        self.state_noise = GaussianState(0.0, self.W)
        self.id = np.zeros([1, self.C.shape[0]])        

    def _check_valid(self, lambda_predict):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        if np.any((lambda_predict * self.spike_rate_dt) > 1): 
            raise ValueError("Cell exploded!")

    def _obs_prob(self, state):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        Loglambda_predict = self.C * state.mean
        lambda_predict = np.exp(Loglambda_predict)/self.spike_rate_dt

        nan_inds = np.isnan(lambda_predict)
        lambda_predict[nan_inds] = 0

        # check max rate is less than 1 b/c it's a probability
        rate_too_high_inds = ((lambda_predict * self.spike_rate_dt) > 1)
        lambda_predict[rate_too_high_inds] = 1./self.spike_rate_dt

        # check min rate is > 0
        rate_too_low_inds = (lambda_predict < 0)
        lambda_predict[rate_too_low_inds] = 0

        invalid_inds = nan_inds | rate_too_high_inds | rate_too_low_inds
        if np.any(invalid_inds):
            pass
            #print np.nonzero(invalid_inds.ravel()[0])
        return lambda_predict

    def _forward_infer(self, st, obs_t, Bu=None, u=None, x_target=None, F=None, obs_is_control_independent=False, **kwargs):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        if np.any(obs_t > 1): 
            raise Exception
        using_control_input = (Bu is not None) or (u is not None) or (x_target is not None)
        if x_target is not None:
            x_target = np.mat(x_target[:,0].reshape(-1,1))
        target_state = x_target

        obs_t = np.mat(obs_t.reshape(-1,1))
        C = self.C
        n_obs, n_states = C.shape
        
        dt = self.spike_rate_dt
        inds, = np.nonzero(self.is_stochastic)
        mesh = np.ix_(inds, inds)
        A = self.A
        W = self.W
        C = C[:,inds]


        # print np.array(x_target).ravel()
        pred_state = self._ssm_pred(st, target_state=x_target, Bu=Bu, u=u, F=F)
        x_pred, P_pred = pred_state.mean, pred_state.cov
        P_pred = P_pred[mesh]

        Loglambda_predict = self.C * x_pred 
        exp = np.vectorize(lambda x: np.real(cmath.exp(x)))
        lambda_predict = exp(np.array(Loglambda_predict).ravel())/dt

        Q_inv = np.mat(np.diag(lambda_predict*dt))

        if np.linalg.cond(P_pred) > 1e5:
            P_est = P_pred;
        else:
            P_est = (P_pred.I + C.T*np.mat(np.diag(lambda_predict*dt))*C).I

        # inflate P_est
        P_est_full = np.mat(np.zeros([n_states, n_states]))
        P_est_full[mesh] = P_est
        P_est = P_est_full 

        unpred_spikes = obs_t - np.mat(lambda_predict*dt).reshape(-1,1)

        x_est = np.mat(np.zeros([n_states, 1]))
        x_est = x_pred + P_est*self.C.T*unpred_spikes
        self.neural_push = P_est*self.C.T*obs_t
        self.P_est = P_est
        post_state = GaussianState(x_est, P_est)
        return post_state

    def __getstate__(self):
        '''
        Return model parameters to be pickled. Overrides the default __getstate__ so that things like the P matrix aren't pickled.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        dict
        '''
        return dict(A=self.A, W=self.W, C=self.C, dt=self.dt, B=self.B, 
                    is_stochastic=self.is_stochastic)

    def tomlab(self, unit_scale=1.):
        '''
        Convert to the MATLAB beta matrix convention from the one used here (different state order, transposed)
        '''
        return np.array(np.hstack([self.C[:,-1], unit_scale*self.C[:,self.is_stochastic]])).T

    @classmethod
    def frommlab(self, beta_mat):
        '''
        Convert from the MATLAB beta matrix convention to the one used here (different state order, transposed)
        '''
        return np.vstack([beta_mat[1:,:], beta_mat[0,:]]).T

    @classmethod
    def MLE_obs_model(cls, hidden_state, obs, include_offset=True, drives_obs=None):
        """
        Unconstrained ML estimator of {C, } given observations and
        the corresponding hidden states
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------            
        """
        assert hidden_state.shape[1] == obs.shape[1]
    
        if isinstance(hidden_state, np.ma.core.MaskedArray):
            mask = ~hidden_state.mask[0,:] # NOTE THE INVERTER 
            inds = np.nonzero([ mask[k]*mask[k+1] for k in range(len(mask)-1)])[0]
    
            X = np.mat(hidden_state[:,mask])
            T = len(np.nonzero(mask)[0])
    
            Y = np.mat(obs[:,mask])
            if include_offset:
                X = np.vstack([ X, np.ones([1,T]) ])
        else:
            num_hidden_state, T = hidden_state.shape
            X = np.mat(hidden_state)
            if include_offset:
                X = np.vstack([ X, np.ones([1,T]) ])
                if not drives_obs == None:
                    drives_obs = np.hstack([drives_obs, True])
                
            Y = np.mat(obs)
        
        X = np.array(X)
        if not drives_obs == None:
            X = X[drives_obs, :]
        Y = np.array(Y)

        # ML estimate of C and Q
        n_units = Y.shape[0]
        n_states = X.shape[0]
        C = np.zeros([n_units, n_states])
        pvalues = np.zeros([n_units, n_states])
        import statsmodels.api as sm
        glm_family = sm.families.Poisson()
        for k in range(n_units):
            model = sm.GLM(Y[k,:], X.T, family=glm_family)
            try:
                model_fit = model.fit()
                C[k,:] = model_fit.params
                pvalues[k,:] = model_fit.pvalues                
            except:
                pvalues[k,:] = np.nan

        return C, pvalues



class OneStepMPCPointProcessFilter(PointProcessFilter):
    '''
    Use MPC with a horizon of 1 to predict 
    '''
    attrs_to_pickle = ['A', 'W', 'C']
    def _pickle_init(self):
        super(OneStepMPCPointProcessFilter, self)._pickle_init()

        self.prev_obs = None
        if not hasattr(self, 'mpc_cost_step'):
            mpc_cost_half_life = 1200.
            batch_time = 0.1
            self.mpc_cost_step = np.exp(np.log(0.5) / (mpc_cost_half_life/batch_time))
            self.ESS = 1000

    def _ssm_pred(self, state, u=None, Bu=None, target_state=None, F=None):
        ''' Docstring
        Run the "predict" step of the Kalman filter/HMM inference algorithm:
            x_{t+1|t} = N(Ax_{t|t}, AP_{t|t}A.T + W)

        Parameters
        ----------
        state: GaussianState instance
            State estimate and estimator covariance of current state
        u: np.mat 
        

        Returns
        -------
        GaussianState instance
            Represents the mean and estimator covariance of the new state estimate
        '''
        A = self.A

        dt = self.dt

        Loglambda_predict = self.C * state.mean 
        exp = np.vectorize(lambda x: np.real(cmath.exp(x)))
        lambda_predict = exp(np.array(Loglambda_predict).ravel())/dt

        Q_inv = np.mat(np.diag(lambda_predict*dt))

        if self.prev_obs is not None:
            y_ref = self.prev_obs
            G = self.C.T * Q_inv
            D = G * self.C
            D[:,-1] = 0
            D[-1,:] = 0

            # Solve for R
            R = 200*D
            
            alpha = A*state
            v = np.linalg.pinv(R + D)*(G*y_ref - D*alpha.mean)
        else:
            v = np.zeros_like(state.mean)

        if Bu is not None:
            return A*state + Bu + self.state_noise + v
        elif u is not None:
            Bu = self.B * u
            return A*state + Bu + self.state_noise + v
        elif target_state is not None:
            B = self.B
            F = self.F
            return (A - B*F)*state + B*F*target_state + self.state_noise  + v
        else:
            return A*state + self.state_noise + v

    def _forward_infer(self, st, obs_t, **kwargs):
        res = super(OneStepMPCPointProcessFilter, self)._forward_infer(st, obs_t, **kwargs)

        # if not (self.prev_obs is None):
        #     # Update the sufficient statistics for the R matrix
        #     l = self.mpc_cost_step
        #     self.E00 = l*self.E00 + (1-l)*(self.prev_obs - self.C*self.A*st.mean)*(self.prev_obs - self.C*self.A*st.mean).T
        #     self.E01 = l*self.E01 + (1-l)*(self.prev_obs - self.C*self.A*st.mean)*(obs_t - self.C*self.A*st.mean).T

        self.prev_obs = obs_t
        return res    

class OneStepMPCPointProcessFilterCovFb(OneStepMPCPointProcessFilter):
    def _ssm_pred(self, state, **kwargs):
        ''' Docstring
        Run the "predict" step of the Kalman filter/HMM inference algorithm:
            x_{t+1|t} = N(Ax_{t|t}, AP_{t|t}A.T + W)

        Parameters
        ----------
        state: GaussianState instance
            State estimate and estimator covariance of current state
        u: np.mat 
        

        Returns
        -------
        GaussianState instance
            Represents the mean and estimator covariance of the new state estimate
        '''
        A = self.A

        dt = self.dt

        Loglambda_predict = self.C * state.mean 
        exp = np.vectorize(lambda x: np.real(cmath.exp(x)))
        lambda_predict = exp(np.array(Loglambda_predict).ravel())/dt

        Q_inv = np.mat(np.diag(lambda_predict*dt))

        from .bmi import GaussianState
        if (self.prev_obs is not None) and (self.r_scale < np.inf):
            y_ref = self.prev_obs
            G = self.C.T * Q_inv
            D = G * self.C
            D[:,-1] = 0
            D[-1,:] = 0

            # Solve for R
            R = self.r_scale*D
            
            alpha = A*state
            v = np.linalg.pinv(R + D)*(G*y_ref - D*alpha.mean)
            I = np.mat(np.eye(D.shape[0]))
            C = self.C
            A = (I - G*C) * self.A
            mean = A*state.mean + G*y_ref
            cov = A*state.cov*A.T + self.W

            return GaussianState(mean, cov)
        else:
            return A*state + self.state_noise
            # v = np.zeros_like(state.mean)



class PPFDecoder(bmi.BMI, bmi.Decoder):
    def __call__(self, obs_t, **kwargs):
        '''
        see bmi.Decoder.__call__ for docs
        '''
        # The PPF model predicts that at most one spike can be observed in 
        # each bin; if more are observed, squash the counts
        # (make a copy of the observation matrix prior to squashing)
        obs_t = obs_t.copy()
        obs_t[obs_t > 1] = 1
        return super(PPFDecoder, self).__call__(obs_t, **kwargs)

    def shuffle(self):
        '''
        Shuffle the neural model
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        '''
        import random
        inds = list(range(self.filt.C.shape[0]))
        random.shuffle(inds)

        # shuffle rows of C
        self.filt.C = self.filt.C[inds, :]

    def compute_suff_stats(self, hidden_state, obs, include_offset=True):
        '''
        Calculate initial estimates of the parameter sufficient statistics used in the RML update rules

        Parameters
        ----------
        hidden_state : np.ndarray of shape (n_states, n_samples)
            Examples of the hidden state x_t taken from training seed data.  
        obs : np.ndarray of shape (n_features, n_samples)
            Multiple neural observations paired with each of the hidden state examples
        include_offset : bool, optional
            If true, a state of all 1's is added to the hidden_state to represent mean offsets. True by default

        Returns
        -------
        R : np.ndarray of shape (n_states, n_states)
            Proportional to covariance of the hidden state samples 
        S : np.ndarray of shape (n_features, n_states)
            Proportional to cross-covariance between 
        T : np.ndarray of shape (n_features, n_features)
            Proportional to covariance of the neural observations
        ESS : float
            Effective number of samples. In the initialization, this is just the 
            dimension of the array passed in, but the parameter can become non-integer 
            during the update procedure as old parameters are "forgotten".
        '''
        n_obs = obs.shape[0]
        nS = hidden_state.shape[0]

        H = np.zeros([n_obs, nS, nS])
        M = np.zeros([n_obs, nS])
        S = np.zeros([n_obs, nS])

        C = self.filt.C[:,self.drives_neurons]

        X = np.array(hidden_state)
        T = X.shape[1]
        if include_offset:
            if not np.all(X[-1,:] == 1):
                X = np.vstack([ X, np.ones([1,T]) ])

        for k in range(n_obs):
            Mu = np.exp(np.dot(C[k,:], X)).ravel()
            Y = obs[k,:]
            H[k] = np.dot((np.tile(np.array(-Mu), [nS, 1]) * X), X.T)
            M[k] = np.dot(Mu, X.T)
            S[k] = np.dot(Y, X.T)

        self.H = H
        self.S = S
        self.M = M

    @property 
    def n_features(self):
        return self.filt.C.shape[0]



'''
Module for the core "features" that can be used to extend and customize a 
task/experiment by multiple inheritance.
'''

from features.debug_features import Profiler, OnlineAnalysis
from features.laser_features import QwalorLaser, MultiQwalorLaser, SwitchedQwalorLaser, LaserState
from features.quattrocento_features import QuattBMI
from riglib.stereo_opengl.window import WindowWithExperimenterDisplay, Window2D
from riglib.stereo_opengl.openxr import WindowVR
from .generator_features import Autostart, AdaptiveGenerator, IgnoreCorrectness, PoissonWait, RandomDelay, Progressbar_fixation
from .peripheral_device_features import Button, Joystick, DualJoystick, Joystick_plus_TouchSensor, KeyboardControl, MouseControl, ForceControl
from .reward_features import RewardSystem, TTLReward, JuiceLogging, PelletReward, JackpotRewards, ProgressBar, TrackingRewards, RewardAudio, PenaltyAudio, ScoreRewards, ConsecutiveJackpot
from .eyetracker_features import EyeData, CalibratedEyeData, PupilLabStreaming, SimulatedEyeData, FixationStart, EyeConstrained, EyeCalibration, EyeStreaming
from .phasespace_features import MotionData, MotionSimulate, MotionAutoAlign
from .optitrack_features import Optitrack, SpheresToCylinders
from .plexon_features import PlexonBMI, RelayPlexon, RelayPlexByte
from .hdf_features import SaveHDF
from .video_recording_features import SingleChannelVideo, E3Video
from .bmi_task_features import NormFiringRates
from .arduino_features import PlexonSerialDIORowByte
from .blackrock_features import BlackrockBMI
from .blackrock_features import RelayBlackrockByte
from .ecube_features import EcubeFileBMI, EcubeBMI, RecordECube
from .sync_features import ArduinoSync, CursorAnalogOut, ScreenSync, HDFSync
from .neuropixel_features import RecordNeuropixels
from .clda_features import CLDA_KFRML_IntendedVelocity

built_in_features = dict(
    keyboard=KeyboardControl,
    mouse=MouseControl,
    optitrack=Optitrack,
    reward_system=RewardSystem, 
    pellet_reward=PelletReward,
    reward_audio=RewardAudio,
    penalty_audio=PenaltyAudio,
    reward_score=ScoreRewards,
    saveHDF=SaveHDF,
    syncHDF=HDFSync,
    autostart=Autostart,
    rand_delay=RandomDelay,
    poisson_wait=PoissonWait,
    window2D=Window2D,
    windowVR=WindowVR,
    cylinder_center_out=SpheresToCylinders,
    adaptive_generator=AdaptiveGenerator,
    button=Button,
    ignore_correctness=IgnoreCorrectness,
    eye_data=EyeData,
    joystick=Joystick,
    dual_joystick=DualJoystick,
    joystick_and_touch = Joystick_plus_TouchSensor,
    calibrated_eye=CalibratedEyeData,
    eye_simulate=SimulatedEyeData,
    fixation_start=FixationStart,
    motion_data=MotionData,
    motion_simulate=MotionSimulate,
    motion_autoalign=MotionAutoAlign,
    norm_firingrates=NormFiringRates,
    jackpot_rewards=JackpotRewards,
    ttl_reward=TTLReward,
    juice_log=JuiceLogging,
    single_video=SingleChannelVideo,
    exp_display=WindowWithExperimenterDisplay,
    relay_arduino=PlexonSerialDIORowByte,
    plexonbmi=PlexonBMI,
    relay_plexon=RelayPlexon,
    relay_plexbyte=RelayPlexByte,
    blackrockbmi        = BlackrockBMI,
    relay_blackrockbyte = RelayBlackrockByte,
    ecube_playback_bmi = EcubeFileBMI,
    ecube_bmi = EcubeBMI,
    ecube_start = RecordECube,
    quatt_bmi = QuattBMI,
    qwalor_laser = QwalorLaser,
    multi_qwalor_laser = MultiQwalorLaser,
    switched_qwalor_laser = SwitchedQwalorLaser,
    laser_state = LaserState,
    e3video = E3Video,
    debug = Profiler,
    online_analysis = OnlineAnalysis,
    arduino_sync=ArduinoSync,
    screen_sync=ScreenSync,
    cursor_sync=CursorAnalogOut,
    progress_bar=ProgressBar,
    tracking_rewards=TrackingRewards,
    neuropixel=RecordNeuropixels,
    eye_streaming=EyeStreaming,
    pupillab_streaming=PupilLabStreaming,
    eye_constrained=EyeConstrained,
    eye_calibration=EyeCalibration, 
    force_sensor=ForceControl,
    show_fixation_progress=Progressbar_fixation,
    clda_kfrml=CLDA_KFRML_IntendedVelocity,
    consecutive_jackpot=ConsecutiveJackpot
)

# >>> features.built_in_features['autostart'].__module__
# 'features.generator_features'
# >>> features.built_in_features['autostart'].__qualname__
# 'Autostart'


'''
Features to include when using the Arduino board to remotely start neural
recording for Plexon/Blackrock/TDT systems and also synchronize data between the task and the neural recordings
'''
import time
import numpy as np
import os
from riglib import serial_dio
import glob
import datetime
import serial

sec_per_min = 60
baudrate = 115200 #9600

class SerialDIORowByte(object):
    '''
    Sends the full data from eyetracking and motiontracking systems directly into Plexon
    '''
    def __init__(self, *args, **kwargs):
        '''
        Constructor for SerialDIORowByte

        Parameters
        ----------
        None

        Returns
        -------
        SerialDIORowByte instance
        '''
        super(SerialDIORowByte, self).__init__(*args, **kwargs)

        self.checked_for_file_changes = False
        self.task_start_time = time.time()
        self.possible_filesizes = []
        self.possible_filenames = []

    def filter_files(self, file_names, start_time=datetime.datetime.today() - datetime.timedelta(days=1)):
        '''
        Filter a list of filenames to find the ones which were timestamped after a particular start_time

        Parameters
        ----------
        file_names : iterable
            Each element should be a string of a filename which exists
        start_time : datetime.datetime object, optional, default = 1 day ago
            Cutoff time for the file timestamp. The file needs
            to be modified after this time in order to pass the filter.
        '''
        return [fname for fname in file_names if datetime.datetime.fromtimestamp(os.stat(fname).st_mtime) > start_time]

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' sets up the NIDAQ card as a sink
        '''
        from riglib import sink
        sink_manager = sink.SinkManager.get_instance()
        self.nidaq = sink_manager.start(serial_dio.SendRowByte)
        self.dbq_kwargs = dict()
        super(SerialDIORowByte, self).init()

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running.
        See riglib.experiment.Experiment.run(). This 'run' method stops the NIDAQ sink after the FSM has stopped running.
        '''
        try:
            super(SerialDIORowByte, self).run()
        finally:
            # Stop the NIDAQ sink
            self.nidaq.stop()

    def set_state(self, condition, **kwargs):
        '''
        Extension of riglib.experiment.Experiment.set_state. Send the name of the next state to
        plexon system and then proceed to the upstream set_state tasks.

        Parameters
        ----------
        condition : string
            Name of new state.
        **kwargs : dict
            Passed to 'super' set_state function

        Returns
        -------
        None
        '''
        self.nidaq.sendMsg(condition)
        super(SerialDIORowByte, self).set_state(condition, **kwargs)

    def _cycle(self):
        if not self.checked_for_file_changes and (time.time() - self.task_start_time > 5):
            filesizes = np.array([os.stat(fname).st_size for fname in self.possible_filenames])
            inds, = np.nonzero(filesizes - self.possible_filesizes)
            if len(inds) == 0:
                print("\n\n\nFile recording did not start? no files have changed size!\n")
            else:
                print("%d neural files have changed since the start of the block" % len(inds))
            self.checked_for_file_changes = True
        super(SerialDIORowByte, self)._cycle()

    def cleanup(self, database, saveid, **kwargs):
        '''
        Function to run at 'cleanup' time, after the FSM has finished executing. See riglib.experiment.Experiment.cleanup
        This 'cleanup' method remotely stops the plexon file recording and then links the file created to the database ID for the current TaskEntry
        '''

        # write the 'stop' command to the port just to be more sure that neural recording has finished.
        port = serial.Serial('/dev/arduino_neurosync', baudrate=baudrate)
        port.write("p")
        super(SerialDIORowByte, self).cleanup(database, saveid, **kwargs)

        # Sleep time so that the neural recording system has time to save cleanly
        time.sleep(5)

        print("Beginning neural data file cleanup")
        # specify which database to save to. If you're running from the web interface, this will always pick the 'default' database

        if "dbname" in kwargs:
            self.dbq_kwargs['dbname']=kwargs["dbname"]

        # Call the appropriate functions in the dbq module to actually link the files
        if self.data_files is None or len(self.data_files) == 0:
            print("\tData files not found properly!\n\tThey will have be manually linked using dbq.save_data!\n\n")
        elif isinstance(self.data_files, str):
            database.save_data(self.data_files, self.db_sys_name, saveid, True, False, **self.dbq_kwargs)
        elif np.iterable(self.data_files):
            for df in self.data_files:
                ext = os.path.splitext(df)[1][1:]
                database.save_data(df, self.db_sys_name, saveid, True, False, ext, **self.dbq_kwargs)

    @classmethod
    def pre_init(cls, saveid=None, **kwargs):
        '''
        Run prior to starting the task to remotely start recording from the plexon system
        '''
        if saveid is not None:
            port = serial.Serial('/dev/arduino_neurosync',baudrate=baudrate)
            port.write('p')
            time.sleep(0.5)
            port.write('r')
            time.sleep(3)

            port.close()
            super(SerialDIORowByte, cls).pre_init(saveid=saveid, **kwargs)

class PlexonSerialDIORowByte(SerialDIORowByte):
    db_sys_name = "plexon"
    storage_root = '/storage/plexon/'

    def init(self):
        """
        Find all the plx files created in the last 24 h before calling the parents init method
        """
        # Find all the plexon files modified in the last day
        file_pattern = self.storage_root + "*.plx"
        self.possible_filenames = self.filter_files(glob.glob(file_pattern))
        self.possible_filesizes = np.array([os.stat(fname).st_size for fname in self.possible_filenames])

        super(PlexonSerialDIORowByte, self).init()

    @property
    def data_files(self):
        '''
        Calculates the plexon file that's most likely associated with the current task
        based on the time at which the task ended and the "last modified" time of the
        plexon files located at /storage/plexon/
        '''
        if hasattr(self, '_data_files'):
            return self._data_files
        else:
            ## Ideally, the correct 'data_files' will be the only file whose filesize has changed since the 'init' function ran..
            filesizes = np.array([os.stat(fname).st_size for fname in self.possible_filenames])
            inds, = np.nonzero(filesizes - self.possible_filesizes)
            if len(inds) == 1:
                print("\tonly one plx file changed since the start of the task.")
                self._data_files = self.possible_filenames[inds[0]]
                return self._data_files

            ## Otherwise, try to find a file whose last modified time is within 60 seconds of the task ending; this requires fairly accurate synchronization between the two machines
            if len(self.event_log) < 1:
                self._data_files = None
                return self._data_files

            start = self.event_log[-1][2]
            #files = "/storage/plexon/*.plx"
            files = "/storage/plexon/test*.plx" #Only use files that have not yet been renamed
            files = sorted(glob.glob(files), key=lambda f: abs(os.stat(f).st_mtime - start))

            if len(files) > 0:
                tdiff = os.stat(files[0]).st_mtime - start
                if abs(tdiff) < sec_per_min:
                    print("\tfound plexon file by finding a file with a timestamp within one minute of the last thing in the event log")
                    self._data_files = files[0]
                    return self._data_files

            ## If both methods fail, return None; cleanup should warn the user that they'll have to link the plexon file manually
            self._data_files = None
            return self._data_files


class BlackrockSerialDIORowByte(SerialDIORowByte):
    db_sys_name = "blackrock2"
    storage_root = "/storage/blackrock"

    file_exts = ["*.nev", "*.ns1", "*.ns2", "*.ns3", "*.ns4", "*.ns5", "*.ns6"]

    def init(self):
        self.possible_filenames = []
        for file_ext in self.file_exts:
            file_pattern1 = os.path.join(self.storage_root, "*/", file_ext)
            file_pattern2 = os.path.join(self.storage_root, file_ext)

            self.possible_filenames += self.filter_files(glob.glob(file_pattern1))
            self.possible_filenames += self.filter_files(glob.glob(file_pattern2))

        self.possible_filesizes = np.array([os.stat(fname).st_size for fname in self.possible_filenames])

        super(BlackrockSerialDIORowByte, self).init()

    @property
    def data_files(self):
        if not hasattr(self, "_data_files"):
            ## Ideally, the correct 'data_files' will be the only file whose filesize has changed since the 'init' function ran..
            filesizes = np.array([os.stat(fname).st_size for fname in self.possible_filenames])
            inds, = np.nonzero(filesizes - self.possible_filesizes)

            files_which_changed_size = [self.possible_filenames[k] for k in inds]

            # make sure that all the base filenames are the same
            file_dates = [x.split(".")[0] for x in files_which_changed_size]
            if len(np.unique(file_dates)) <= 1:
                self._data_files = files_which_changed_size
            else:
                print("Filenames ambiguous! BlackrockSerialDIORowByte cannot figure out which files are associated with this block")
                print("these are the options:")
                print(files_which_changed_size)
                print()

        return self._data_files


class TDTSerialDIORowByte(SerialDIORowByte):
    '''
    Sends the full data from eyetracking and motiontracking systems directly into Plexon
    '''
    db_sys_name = "tdt"
    def __init__(self, *args, **kwargs):
        super(TDTSerialDIORowByte, self).__init__(*args, **kwargs)
        self.file_ext = kwargs.pop('file_ext', '.tev')
        self.data_root = kwargs.pop('data_root', '/storage/tdt/')
        self.file_pattern = os.path.join(self.data_root, '*' + self.file_ext)

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' sets up the NIDAQ card as a sink
        '''
        from riglib import sink
        sink_manager = sink.SinkManager.get_instance()
        self.nidaq = sink_manager.start(self.ni_out)
        super(TDTSerialDIORowByte, self).init()

    @property
    def ni_out(self):
        '''
        Specify the output interface; can be overridden in child classes as long as
        this method returns a class which has the same instance methods (close, register, send, sendMsg, etc.)
        '''
        # TODO ni_out ---> iface
        from riglib import serial_dio
        return serial_dio.SendRowByte

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running.
        See riglib.experiment.Experiment.run(). This 'run' method stops the NIDAQ sink after the FSM has stopped running.
        '''
        try:
            super(TDTSerialDIORowByte, self).run()
        finally:
            # Stop the NIDAQ sink
            self.nidaq.stop()

    def set_state(self, condition, **kwargs):
        '''
        Extension of riglib.experiment.Experiment.set_state. Send the name of the next state to
        plexon system and then proceed to the upstream set_state tasks.

        Parameters
        ----------
        condition : string
            Name of new state.
        **kwargs : dict
            Passed to 'super' set_state function

        Returns
        -------
        None
        '''
        self.nidaq.sendMsg(condition)
        super(TDTSerialDIORowByte, self).set_state(condition, **kwargs)

    def cleanup(self, database, saveid, **kwargs):
        '''
        Function to run at 'cleanup' time, after the FSM has finished executing. See riglib.experiment.Experiment.cleanup
        This 'cleanup' method remotely stops the plexon file recording and then links the file created to the database ID for the current TaskEntry
        '''
        # Stop recording
        # import comedi
        import time

        # com = comedi.comedi_open("/dev/comedi0")
        # comedi.comedi_dio_bitfield2(com, 0, 16, 16, 16)

        # port = serial.Serial(glob.glob("/dev/ttyACM*")[0], baudrate=9600)
        # port.write('p')
        # port.close()

        super(TDTSerialDIORowByte, self).cleanup(database, saveid, **kwargs)

        # Sleep time so that the plx file has time to save cleanly

        time.sleep(2)
        """
        dbname = kwargs['dbname'] if 'dbname' in kwargs else 'default'
        if self.plexfile is not None:
            if dbname == 'default':
                database.save_data(self.plexfile, "plexon", saveid, True, False)
            else:
                database.save_data(self.plexfile, "plexon", saveid, True, False, dbname=dbname)
        else:
            print '\n\nTDT file not found properly! It will have to be manually linked!\n\n'
        """

    @classmethod
    def pre_init(cls, saveid=None, **kwargs):
        '''
        Run prior to starting the task to remotely start recording from the plexon system
        '''
        if saveid is not None:
            port = serial.Serial('/dev/arduino_neurosync', baudrate=baudrate)
            # for k in range(5):
            port.write('p')
            time.sleep(0.5)
            port.write('r')
            port.close()

            time.sleep(3)
            super(TDTSerialDIORowByte, cls).pre_init(saveid=saveid, **kwargs)

    @property
    def data_files(self):
        return None



# Generated by Django 2.2.1 on 2019-11-12 23:02

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AutoAlignment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('name', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='Feature',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('visible', models.BooleanField(blank=True, default=True)),
            ],
        ),
        migrations.CreateModel(
            name='Generator',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('params', models.TextField()),
                ('static', models.BooleanField()),
                ('visible', models.BooleanField(blank=True, default=True)),
            ],
        ),
        migrations.CreateModel(
            name='Sequence',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=128)),
                ('params', models.TextField()),
                ('sequence', models.TextField(blank=True)),
                ('generator', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.Generator')),
            ],
        ),
        migrations.CreateModel(
            name='Subject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
            ],
        ),
        migrations.CreateModel(
            name='System',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('path', models.TextField()),
                ('archive', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('visible', models.BooleanField(blank=True, default=True)),
            ],
        ),
        migrations.CreateModel(
            name='TaskEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('params', models.TextField()),
                ('report', models.TextField()),
                ('notes', models.TextField()),
                ('visible', models.BooleanField(blank=True, default=True)),
                ('backup', models.BooleanField(blank=True, default=False)),
                ('feats', models.ManyToManyField(to='tracker.Feature')),
                ('sequence', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='tracker.Sequence')),
                ('subject', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.Subject')),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.Task')),
            ],
        ),
        migrations.AddField(
            model_name='sequence',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.Task'),
        ),
        migrations.CreateModel(
            name='Decoder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=128)),
                ('path', models.TextField()),
                ('entry', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.TaskEntry')),
            ],
        ),
        migrations.CreateModel(
            name='DataFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('local', models.BooleanField(default=True)),
                ('archived', models.BooleanField(default=False)),
                ('path', models.CharField(max_length=256)),
                ('entry', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='tracker.TaskEntry')),
                ('system', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.System')),
            ],
        ),
        migrations.CreateModel(
            name='Calibration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=128)),
                ('params', models.TextField()),
                ('subject', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.Subject')),
                ('system', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.System')),
            ],
        ),
    ]



''' Mock experiment classes for testing '''
from .experiment import LogExperiment, FSMTable, StateTransitions, Sequence

event1to2 = [False, True,  False, False, False, False, False, False]
event1to3 = [False, False, False, False, True,  False, False, False]
event2to3 = [False, False, True,  False, False, False, False, False]
event2to1 = [False, False, False, False, False, False, False, False]
event3to2 = [False, False, False, False, False, True,  False, False]
event3to1 = [False, False, False, True,  False, False, False, False]

class MockLogExperiment(LogExperiment):
    status = FSMTable(
        state1=StateTransitions(event1to2='state2', event1to3='state3'),
        state2=StateTransitions(event2to3='state3', event2to1='state1'),
        state3=StateTransitions(event3to2='state2', event3to1='state1'),
    )
    state = 'state1'

    def __init__(self, *args, **kwargs):
        self.iter_idx = 0
        super(MockLogExperiment, self).__init__(*args, **kwargs)

    def _cycle(self):
        self.iter_idx += 1
        super(MockLogExperiment, self)._cycle()

    def _start_state3(self): pass
    def _while_state3(self): pass
    def _end_state3(self): pass
    def _start_state2(self): pass
    def _while_state2(self): pass
    def _end_state2(self): pass
    def _start_state1(self): pass
    def _while_state1(self): pass
    def _end_state1(self): pass
    ################## State trnasition test functions ##################
    def _test_event3to1(self, time_in_state): return event3to1[self.iter_idx]
    def _test_event3to2(self, time_in_state): return event3to2[self.iter_idx]
    def _test_event2to3(self, time_in_state): return event2to3[self.iter_idx]
    def _test_event2to1(self, time_in_state): return event2to1[self.iter_idx]
    def _test_event1to3(self, time_in_state): return event1to3[self.iter_idx]
    def _test_event1to2(self, time_in_state): return event1to2[self.iter_idx]
    def _test_stop(self, time_in_state):
        return self.iter_idx >= len(event1to2) - 1

    def get_time(self):
        return self.iter_idx



class MockSequence(Sequence):
    event1to2 = [False, True,  False, True, False, True, False, True, False, True, False]
    event1to3 = [False, False, False, False, False,  False, False, False, False, False, False]
    event2to3 = [False, False, False,  False, False, False, False, False, False, False, False]
    event2to1 = [False, False, True, False, True, False, True, False, True, False, False]
    event3to2 = [False, False, False, False, False, False,  False, False, False, False, False]
    event3to1 = [False, False, False, False,  False, False, False, False, False, False, False]

    status = FSMTable(
        wait=StateTransitions(event1to2='state2', event1to3='state3'),
        state2=StateTransitions(event2to3='state3', event2to1='wait'),
        state3=StateTransitions(event3to2='state2', event3to1='wait'),
    )
    state = 'wait'

    def __init__(self, *args, **kwargs):
        self.iter_idx = 0
        self.target_history = []
        super(MockSequence, self).__init__(*args, **kwargs)

    def _cycle(self):
        self.iter_idx += 1
        super(MockSequence, self)._cycle()

    def _start_state3(self): pass
    def _while_state3(self): pass
    def _end_state3(self): pass
    def _start_state2(self): pass
    def _while_state2(self): pass
    def _end_state2(self): pass
    def _start_state1(self): pass
    def _while_state1(self): pass
    def _end_state1(self): pass
    ################## State trnasition test functions ##################
    def _test_event3to1(self, time_in_state): return self.event3to1[self.iter_idx]
    def _test_event3to2(self, time_in_state): return self.event3to2[self.iter_idx]
    def _test_event2to3(self, time_in_state): return self.event2to3[self.iter_idx]
    def _test_event2to1(self, time_in_state): return self.event2to1[self.iter_idx]
    def _test_event1to3(self, time_in_state): return self.event1to3[self.iter_idx]
    def _test_event1to2(self, time_in_state): return self.event1to2[self.iter_idx]
    def _test_stop(self, time_in_state):
        return self.iter_idx >= len(event1to2) - 1

    def get_time(self):
        return self.iter_idx

    def _start_wait(self):
        super(MockSequence, self)._start_wait()
        self.target_history.append(self.next_trial)


class MockSequenceWithGenerators(Sequence):
    status = FSMTable(
        wait=StateTransitions(start_trial="trial"),
        trial=StateTransitions(target_reached="reward"),
        reward=StateTransitions(reward_complete="wait"),
    )
    state = 'wait'

    sequence_generators = ['gen_fn1', 'gen_fn2']

    def __init__(self, *args, **kwargs):
        self.target_history = []
        self.sim_state_seq = []
        for k in range(4):
            self.sim_state_seq += [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 2]

        self.current_state = None
        super().__init__(*args, **kwargs)

    def init(self):
        self.add_dtype("target_state", int, (1,))
        self.add_dtype("current_state", int, (1,))
        super().init()

    def get_time(self):
        return self.cycle_count * 1.0/60

    @staticmethod
    def gen_fn1(n_targets=4):
        target_seq = [1, 2] * n_targets
        return [{"target":x} for x in target_seq]

    @staticmethod
    def gen_fn2(n_targets=4):
        target_seq = [1, 2] * n_targets
        return [{"target":x} for x in target_seq]

    def _test_start_trial(self, ts):
        return True

    def _test_reward_complete(self, ts):
        return True

    def _cycle(self):
        self.current_state = self.sim_state_seq[self.cycle_count % len(self.sim_state_seq)]
        self.task_data["target_state"] = self._gen_target
        self.task_data["current_state"] = self.current_state

        if self.cycle_count == 21:
            self.record_annotation("test annotation")

        super()._cycle()

    def _test_target_reached(self, ts):
        return self.current_state == self._gen_target

from . import traits
class MockSequenceWithTraits(MockSequenceWithGenerators):
    options_trait = traits.OptionsList(["option1", "option2"], desc='options', label="Options")
    float_trait = traits.Float(15, desc="float", label="Float")



import subprocess
from shlex import split
import sys, time
import os

from .control_comedi_swig import comedi_init, set_bits_in_nidaq_using_mask_and_data


def init_comedi():
    '''
    a wrapper that wraps around control_comedi_swig.comedi_init
    to set all channels to output

    '''
    COMEDI_SETUP_STRING = '/dev/comedi0'
    EXPECTED_COMEDI_SETUP_STR = 'found comedi system with 24 channels\n0'

    comedi_setup_str = comedi_init(COMEDI_SETUP_STRING)

    print(comedi_setup_str)

def write_to_comedi(data, mask=255, base_channel=0, method='use_swig', debug=False):
    '''
    this is just a wrapper that can call 
    Inputs:
        data[int]: 0 o 255 integer number  that writes to the ports masked by the mask
        mask[int]: the set of ports to write to
        base_channel[int]: the channel offset to which the bits are written, default to 0
        method[str]: default to use 'use_swig'. can also choose 'use_subprocess'

    Output
        retval[int]: 0 means properly set to the ports

    example usage:
    init_comedi()
    #set the the first 4 bits to be 1 as 15 = 0b1111
    write_to_comedi(15)
    '''
    
    if method == 'use_swig':
        return write_to_comedi_swig(data, mask = mask, base_channel=base_channel,
                             debug = debug)
    elif method == 'subprocess':
        return write_to_comedi_subprocess(data, mask = mask, debug = debug)
    else:
        raise Exception(f'unrecognized method {method}')
    

def write_to_comedi_swig(data, mask=255, base_channel=0, debug=False):
    '''
    this function requires the C interface code generated by swig interface
    Inputs:
        data[int]: 0 to 255 that maps to the 8 bit channels
        mask[int]: that mask which channels to write e.g. 0x3 means 0b11 writes to channels 0 and 1
        base_channel[int]: the channel offset to which the bits are written, default to 0
        debug: debug flag

    Output
        retval[int]: 0 means succussfully set, -1 means not
    '''
    retval = set_bits_in_nidaq_using_mask_and_data(mask, data, base_channel)

    return retval


COMEDI_DIO_WRITE = 0 #this writes to a channel
COMEDI_DIO_BITFIELD = 1 #tis writes to multiple channel
MODULE_FOLDER_PATH = os.path.dirname(__file__)

def write_to_comedi_subprocess(data, mask=255, debug=False):
    '''
    data: an integer bet. 0 and 255 (inclusive) that maps to the 8 bit channels. 
    mask: an integer that mask which channels to write e.g. 0x3 means 0b11 writes to channels 0 and 1
    debug: decoding switch, if
        True: print out the command
        False (default): do nothing. 
    eg. write_to_comedi(3) sets first two channels to high. 

    note: for efficiency, this function assumes data is in integer in [0, 255], 

    '''
    
    cmd = f'{MODULE_FOLDER_PATH}/control_comedi {COMEDI_DIO_BITFIELD} {hex(mask)} {hex(data)}'

    if debug: print(cmd)

    proc = subprocess.Popen(split(cmd), 
                        stderr= subprocess.PIPE,
                        stdout= subprocess.PIPE,
                        stdin= subprocess.PIPE,
                        shell = False)
    retval = (proc.stdout.readline())
    proc.stdout.flush()
    return retval


def check_comedi() -> bool:
    '''
    this function checks whether 
    comedi is working by simply checking the return has the comedi string in it
    it is knd of useless
    '''
    cmd = 'cat /proc/comedi'
    
    retval = subprocess_write(cmd, debug=True)

    if 'comedi' in retval.decode("utf-8"): 
        return True
    else: 
        return False

def time_write_to_comedi():
    '''
    this function calculates how long one write takes, 
    but does not test how long the signal appears on the port. 
    return the time difference in s
    '''
    data = 255

    t1 = time.perf_counter()
    write_to_comedi(data, mask = 255, debug = True)
    t2 = time.perf_counter()

    return t2 - t1

def subprocess_write(cmd:str, debug:bool = False) -> bytes:
    '''
    helper function that writes a command and returns a string
    note that it only returns the first line
    '''
    if debug: print(f'{__name__}: system command: {cmd}')

    proc = subprocess.Popen(split(cmd), 
                        stderr= subprocess.PIPE,
                        stdout= subprocess.PIPE,
                        stdin= subprocess.PIPE,
                        shell = False)
    retval = (proc.stdout.readline())

    if debug: print(f'{__name__}: one line received: {retval}')

    proc.stdout.flush()

    return retval


if __name__ == "__main__":

    print(f'calling c process takes: {time_write_to_comedi()} s')

